# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import mock
import pytest

from hikari import guilds
from hikari import presences
from hikari import snowflakes
from hikari.events import guild_events
from tests.hikari import hikari_test_helpers


class TestGuildEvent:
    @pytest.fixture
    def event(self):
        cls = hikari_test_helpers.mock_class_namespace(
            guild_events.GuildEvent, guild_id=mock.PropertyMock(return_value=snowflakes.Snowflake(534123123))
        )
        return cls()

    def test_get_guild_when_available(self, event):
        result = event.get_guild()

        assert result is event.app.cache.get_available_guild.return_value
        event.app.cache.get_available_guild.assert_called_once_with(534123123)
        event.app.cache.get_unavailable_guild.assert_not_called()

    def test_get_guild_when_unavailable(self, event):
        event.app.cache.get_available_guild.return_value = None
        result = event.get_guild()

        assert result is event.app.cache.get_unavailable_guild.return_value
        event.app.cache.get_unavailable_guild.assert_called_once_with(534123123)
        event.app.cache.get_available_guild.assert_called_once_with(534123123)

    def test_get_guild_cacheless(self, event):
        event = hikari_test_helpers.mock_class_namespace(guild_events.GuildEvent, app=object())()

        assert event.get_guild() is None

    @pytest.mark.asyncio
    async def test_fetch_guild(self, event):
        event.app.rest.fetch_guild = mock.AsyncMock()
        result = await event.fetch_guild()

        assert result is event.app.rest.fetch_guild.return_value
        event.app.rest.fetch_guild.assert_called_once_with(534123123)

    @pytest.mark.asyncio
    async def test_fetch_guild_preview(self, event):
        event.app.rest.fetch_guild_preview = mock.AsyncMock()
        result = await event.fetch_guild_preview()

        assert result is event.app.rest.fetch_guild_preview.return_value
        event.app.rest.fetch_guild_preview.assert_called_once_with(534123123)


class TestGuildAvailableEvent:
    @pytest.fixture
    def event(self):
        return guild_events.GuildAvailableEvent(
            shard=object(),
            guild=mock.Mock(guilds.Guild),
            emojis={},
            stickers={},
            roles={},
            channels={},
            members={},
            presences={},
            voice_states={},
            threads={},
        )

    def test_app_property(self, event):
        assert event.app is event.guild.app

    def test_guild_id_property(self, event):
        event.guild.id = 123
        assert event.guild_id == 123


class TestGuildUpdateEvent:
    @pytest.fixture
    def event(self):
        return guild_events.GuildUpdateEvent(
            shard=object(),
            guild=mock.Mock(guilds.Guild),
            old_guild=mock.Mock(guilds.Guild),
            emojis={},
            stickers={},
            roles={},
        )

    def test_app_property(self, event):
        assert event.app is event.guild.app

    def test_guild_id_property(self, event):
        event.guild.id = 123
        assert event.guild_id == 123

    def test_old_guild_id_property(self, event):
        event.old_guild.id = 123
        assert event.old_guild.id == 123


class TestBanEvent:
    @pytest.fixture
    def event(self):
        return hikari_test_helpers.mock_class_namespace(guild_events.BanEvent)()

    def test_app_property(self, event):
        assert event.app is event.user.app


class TestPresenceUpdateEvent:
    @pytest.fixture
    def event(self):
        return guild_events.PresenceUpdateEvent(
            shard=object(),
            presence=mock.Mock(presences.MemberPresence),
            old_presence=mock.Mock(presences.MemberPresence),
            user=mock.Mock(),
        )

    def test_app_property(self, event):
        assert event.app is event.presence.app

    def test_user_id_property(self, event):
        event.presence.user_id = 123
        assert event.user_id == 123

    def test_guild_id_property(self, event):
        event.presence.guild_id = 123
        assert event.guild_id == 123

    def test_old_presence(self, event):
        event.old_presence.id = 123
        event.old_presence.guild_id = 456

        assert event.old_presence.id == 123
        assert event.old_presence.guild_id == 456


class TestGuildStickersUpdateEvent:
    @pytest.fixture
    def event(self):
        return guild_events.StickersUpdateEvent(
            app=mock.Mock(),
            shard=mock.Mock(),
            guild_id=snowflakes.Snowflake(690),
            old_stickers=(mock.Mock(), mock.Mock()),
            stickers=(mock.Mock(), mock.Mock(), mock.Mock()),
        )

    @pytest.mark.asyncio
    async def test_fetch_stickers(self, event):
        event.app.rest.fetch_guild_stickers = mock.AsyncMock()

        assert await event.fetch_stickers() is event.app.rest.fetch_guild_stickers.return_value

        event.app.rest.fetch_guild_stickers.assert_awaited_once_with(event.guild_id)


class TestAuditLogEntryCreateEvent:
    @pytest.fixture
    def event(self):
        return guild_events.AuditLogEntryCreateEvent(shard=mock.Mock(), entry=mock.Mock())

    def test_app_property(self, event):
        assert event.app is event.entry.app

    def test_guild_id_property(self, event):
        assert event.guild_id is event.entry.guild_id



# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Various utilities that may be used in a cache-impl."""

from __future__ import annotations

__all__: typing.Sequence[str] = (
    "BaseData",
    "Cache3DMappingView",
    "CacheMappingView",
    "DataT",
    "EmptyCacheView",
    "GuildRecord",
    "InviteData",
    "KeyT",
    "KnownCustomEmojiData",
    "MemberData",
    "MemberPresenceData",
    "MessageData",
    "RefCell",
    "RichActivityData",
    "ValueT",
    "VoiceStateData",
    "copy_guild_channel",
    "unwrap_ref_cell",
)

import abc
import copy
import typing

import attrs

from hikari import embeds as embeds_
from hikari import emojis
from hikari import guilds
from hikari import invites
from hikari import messages
from hikari import presences
from hikari import snowflakes
from hikari import stickers as stickers_
from hikari import undefined
from hikari import voices
from hikari.api import cache
from hikari.internal import attrs_extensions
from hikari.internal import collections

if not typing.TYPE_CHECKING:
    # This is insanely hacky, but it is needed for ruff to not complain until it gets type inference
    from hikari.internal import typing_extensions


if typing.TYPE_CHECKING:
    import datetime

    import typing_extensions  # noqa: TC004
    from typing_extensions import Self

    from hikari import applications
    from hikari import channels as channels_
    from hikari import components as components_
    from hikari import polls as polls_
    from hikari import traits
    from hikari import users as users_
    from hikari.interactions import base_interactions

ChannelT = typing.TypeVar("ChannelT", bound="channels_.PermissibleGuildChannel")
DataT = typing.TypeVar("DataT", bound="BaseData[typing.Any]")
"""Type-hint for "data" objects used for storing and building entities."""
KeyT = typing.TypeVar("KeyT", bound=typing.Hashable)
"""Type-hint for mapping keys."""
ValueT = typing.TypeVar("ValueT")
"""Type-hint for mapping values."""


class CacheMappingView(cache.CacheView[KeyT, ValueT]):
    """A cache mapping view implementation used for representing cached data.

    Parameters
    ----------
    items
        A mapping of keys to the values in their raw forms, wrapped by a ref
        wrapper or in a data form.
    builder
        The callable used to build entities before they're returned by the
        mapping. This is used to cover the case when items stores [`DataT`][] objects.
    """

    __slots__: typing.Sequence[str] = ("_builder", "_data")

    @typing.overload
    def __init__(self, items: typing.Mapping[KeyT, ValueT]) -> None: ...

    @typing.overload
    def __init__(self, items: typing.Mapping[KeyT, DataT], *, builder: typing.Callable[[DataT], ValueT]) -> None: ...

    def __init__(
        self,
        items: typing.Mapping[KeyT, ValueT] | typing.Mapping[KeyT, DataT],
        *,
        builder: typing.Callable[[DataT], ValueT] | None = None,
    ) -> None:
        self._builder = builder
        self._data = items

    @staticmethod
    def _copy(value: ValueT) -> ValueT:
        return copy.copy(value)

    @typing_extensions.override
    def __contains__(self, key: object) -> bool:
        return key in self._data

    @typing_extensions.override
    def __getitem__(self, key: KeyT) -> ValueT:
        entry = self._data[key]

        if self._builder:
            return self._builder(entry)  # type: ignore[arg-type]

        return self._copy(entry)  # type: ignore[arg-type]

    @typing_extensions.override
    def __iter__(self) -> typing.Iterator[KeyT]:
        return iter(self._data)

    @typing_extensions.override
    def __len__(self) -> int:
        return len(self._data)

    @typing.overload
    def get_item_at(self, index: int, /) -> ValueT: ...

    @typing.overload
    def get_item_at(self, index: slice, /) -> typing.Sequence[ValueT]: ...

    @typing_extensions.override
    def get_item_at(self, index: slice | int, /) -> ValueT | typing.Sequence[ValueT]:
        return collections.get_index_or_slice(self, index)


class EmptyCacheView(cache.CacheView[typing.Any, typing.Any]):
    """An empty cache view implementation."""

    __slots__: typing.Sequence[str] = ()

    @typing_extensions.override
    def __contains__(self, _: object) -> typing.Literal[False]:
        return False

    @typing_extensions.override
    def __getitem__(self, key: object) -> typing.NoReturn:
        raise KeyError(key)

    @typing_extensions.override
    def __iter__(self) -> typing.Iterator[typing.Any]:
        yield from ()

    @typing_extensions.override
    def __len__(self) -> typing.Literal[0]:
        return 0

    @typing_extensions.override
    def get_item_at(self, index: slice | int) -> typing.NoReturn:
        raise IndexError(index)


@attrs_extensions.with_copy
@attrs.define(repr=False, weakref_slot=False)
class GuildRecord:
    """An object used for storing guild specific cached information in-memory.

    This includes references to the cached entities that "belong" to the guild
    by ID if it's globally unique or by object if it's only unique within the
    guild.
    """

    is_available: bool | None = attrs.field(default=None)
    """Whether the cached guild is available or not.

    This will be [`None`][] when no `guild` is also
    [`None`][] else [`bool`][].
    """

    guild: guilds.GatewayGuild | None = attrs.field(default=None)
    """A cached guild object.

    This will be [`None`][] if not cached.
    """

    channels: typing.MutableSet[snowflakes.Snowflake] | None = attrs.field(default=None)
    """A set of the IDs of the guild channels cached for this guild.

    This will be [`None`][] if no channels are cached for this guild.
    """

    threads: typing.MutableSet[snowflakes.Snowflake] | None = attrs.field(default=None)
    """A set of the IDs of the guild threads cached for this guild.

    This will be [`None`][] if no threads are cached for this guild.
    """

    emojis: typing.MutableSet[snowflakes.Snowflake] | None = attrs.field(default=None)
    """A set of the IDs of the emojis cached for this guild.

    This will be [`None`][] if no emojis are cached for this guild.
    """

    stickers: typing.MutableSet[snowflakes.Snowflake] | None = attrs.field(default=None)
    """A sequence of sticker IDs cached for this guild.

    This will be [`None`][] if no stickers are cached for this guild.
    """

    invites: typing.MutableSequence[str] | None = attrs.field(default=None)
    """A set of the [`str`][] codes of the invites cached for this guild.

    This will be [`None`][] if no invites are cached for this guild.
    """

    members: collections.ExtendedMutableMapping[snowflakes.Snowflake, RefCell[MemberData]] | None = attrs.field(
        default=None
    )
    """A mapping of user IDs to the objects of members cached for this guild.

    This will be [`None`][] if no members are cached for this guild.
    """

    presences: collections.ExtendedMutableMapping[snowflakes.Snowflake, MemberPresenceData] | None = attrs.field(
        default=None
    )
    """A mapping of user IDs to objects of the presences cached for this guild.

    This will be [`None`][] if no presences are cached for this guild.
    """

    roles: typing.MutableSet[snowflakes.Snowflake] | None = attrs.field(default=None)
    """A set of the IDs of the roles cached for this guild.

    This will be [`None`][] if no roles are cached for this guild.
    """

    voice_states: collections.ExtendedMutableMapping[snowflakes.Snowflake, VoiceStateData] | None = attrs.field(
        default=None
    )
    """A mapping of user IDs to objects of the voice states cached for this guild.

    This will be [`None`][] if no voice states are cached for this guild.
    """

    def empty(self) -> bool:
        """Check whether this guild record has any resources attached to it.

        Returns
        -------
        bool
            Whether this guild record has any resources attached to it.
        """
        # As `.is_available` should be paired with `.guild`, we don't need to check both.
        return not any(
            (
                self.channels,
                self.emojis,
                self.guild,
                self.invites,
                self.members,
                self.presences,
                self.roles,
                self.voice_states,
            )
        )


class BaseData(abc.ABC, typing.Generic[ValueT]):
    """A data class used for in-memory storage of entities in a more primitive form.

    !!! note
        This base implementation assumes that all the fields it'll handle will
        be immutable and to handle mutable fields you'll have to override
        build_entity and build_from_entity to explicitly copy them.
    """

    __slots__: typing.Sequence[str] = ()

    @abc.abstractmethod
    def build_entity(self, app: traits.RESTAware, /) -> ValueT:
        """Build an entity object from this data object.

        Parameters
        ----------
        app
            The hikari application the built object should be bound to.

        Returns
        -------
        ValueT
            The initialised entity object.
        """

    @classmethod
    @abc.abstractmethod
    def build_from_entity(cls, entity: ValueT, /) -> Self:
        """Build a data object from an initialised entity.

        Parameters
        ----------
        entity
            The entity object to build a data class from.

        Returns
        -------
        DataT
            The built data class.
        """


@attrs_extensions.with_copy
@attrs.define(kw_only=True, repr=False, weakref_slot=False)
class InviteData(BaseData[invites.InviteWithMetadata]):
    """A data model for storing invite data in an in-memory cache."""

    code: str = attrs.field()
    guild_id: snowflakes.Snowflake | None = attrs.field()
    channel_id: snowflakes.Snowflake = attrs.field()
    inviter: RefCell[users_.User] | None = attrs.field()
    target_type: invites.TargetType | int | None = attrs.field()
    target_user: RefCell[users_.User] | None = attrs.field()
    target_application: applications.InviteApplication | None = attrs.field()
    uses: int = attrs.field()
    max_uses: int | None = attrs.field()
    max_age: datetime.timedelta | None = attrs.field()
    is_temporary: bool = attrs.field()
    created_at: datetime.datetime = attrs.field()

    @typing_extensions.override
    def build_entity(self, app: traits.RESTAware, /) -> invites.InviteWithMetadata:
        return invites.InviteWithMetadata(
            code=self.code,
            guild_id=self.guild_id,
            channel_id=self.channel_id,
            target_type=self.target_type,
            target_application=self.target_application,
            uses=self.uses,
            max_uses=self.max_uses,
            max_age=self.max_age,
            is_temporary=self.is_temporary,
            created_at=self.created_at,
            approximate_member_count=None,
            approximate_active_member_count=None,
            channel=None,
            guild=None,
            app=app,
            inviter=self.inviter.copy() if self.inviter else None,
            target_user=self.target_user.copy() if self.target_user else None,
            expires_at=self.created_at + self.max_age if self.max_age else None,
        )

    @classmethod
    @typing_extensions.override
    def build_from_entity(
        cls,
        invite: invites.InviteWithMetadata,
        /,
        *,
        inviter: RefCell[users_.User] | None = None,
        target_user: RefCell[users_.User] | None = None,
    ) -> InviteData:
        if not inviter and invite.inviter:
            inviter = RefCell(copy.copy(invite.inviter))

        if not target_user and invite.target_user:
            target_user = RefCell(copy.copy(invite.target_user))

        return cls(
            code=invite.code,
            guild_id=invite.guild_id,
            channel_id=invite.channel_id,
            target_type=invite.target_type,
            target_application=invite.target_application,
            uses=invite.uses,
            max_uses=invite.max_uses,
            max_age=invite.max_age,
            is_temporary=invite.is_temporary,
            created_at=invite.created_at,
            inviter=inviter,
            target_user=target_user,
        )


@attrs_extensions.with_copy
@attrs.define(kw_only=True, repr=False, weakref_slot=False)
class MemberData(BaseData[guilds.Member]):
    """A data model for storing member data in an in-memory cache."""

    user: RefCell[users_.User] = attrs.field()
    guild_id: snowflakes.Snowflake = attrs.field()
    nickname: str | None = attrs.field()
    guild_avatar_decoration: users_.AvatarDecoration | None = attrs.field()
    guild_avatar_hash: str | None = attrs.field()
    guild_banner_hash: str | None = attrs.field()
    role_ids: tuple[snowflakes.Snowflake, ...] = attrs.field()
    joined_at: datetime.datetime | None = attrs.field()
    premium_since: datetime.datetime | None = attrs.field()
    is_deaf: undefined.UndefinedOr[bool] = attrs.field()
    is_mute: undefined.UndefinedOr[bool] = attrs.field()
    is_pending: undefined.UndefinedOr[bool] = attrs.field()
    raw_communication_disabled_until: datetime.datetime | None = attrs.field()
    guild_flags: guilds.GuildMemberFlags | int = attrs.field()
    # meta-attribute
    has_been_deleted: bool = attrs.field(default=False, init=False)

    @classmethod
    @typing_extensions.override
    def build_from_entity(cls, member: guilds.Member, /, *, user: RefCell[users_.User] | None = None) -> MemberData:
        return cls(
            guild_id=member.guild_id,
            nickname=member.nickname,
            joined_at=member.joined_at,
            premium_since=member.premium_since,
            guild_avatar_decoration=member.guild_avatar_decoration,
            guild_avatar_hash=member.guild_avatar_hash,
            guild_banner_hash=member.guild_banner_hash,
            is_deaf=member.is_deaf,
            is_mute=member.is_mute,
            is_pending=member.is_pending,
            user=user or RefCell(copy.copy(member.user)),
            raw_communication_disabled_until=member.raw_communication_disabled_until,
            guild_flags=member.guild_flags,
            # role_ids is a special case as it may be mutable so we want to ensure it's
            # immutable when cached.
            role_ids=tuple(member.role_ids),
        )

    @typing_extensions.override
    def build_entity(self, _: traits.RESTAware, /) -> guilds.Member:
        return guilds.Member(
            guild_id=self.guild_id,
            nickname=self.nickname,
            role_ids=self.role_ids,
            joined_at=self.joined_at,
            guild_avatar_decoration=self.guild_avatar_decoration,
            guild_avatar_hash=self.guild_avatar_hash,
            guild_banner_hash=self.guild_banner_hash,
            premium_since=self.premium_since,
            is_deaf=self.is_deaf,
            is_mute=self.is_mute,
            is_pending=self.is_pending,
            raw_communication_disabled_until=self.raw_communication_disabled_until,
            user=self.user.copy(),
            guild_flags=self.guild_flags,
        )


@attrs_extensions.with_copy
@attrs.define(kw_only=True, repr=False, weakref_slot=False)
class KnownCustomEmojiData(BaseData[emojis.KnownCustomEmoji]):
    """A data model for storing known custom emoji data in an in-memory cache."""

    id: snowflakes.Snowflake = attrs.field()
    name: str = attrs.field()
    is_animated: bool = attrs.field()
    guild_id: snowflakes.Snowflake = attrs.field()
    role_ids: tuple[snowflakes.Snowflake, ...] = attrs.field()
    user: RefCell[users_.User] | None = attrs.field()
    is_colons_required: bool = attrs.field()
    is_managed: bool = attrs.field()
    is_available: bool = attrs.field()

    @classmethod
    @typing_extensions.override
    def build_from_entity(
        cls, emoji: emojis.KnownCustomEmoji, /, *, user: RefCell[users_.User] | None = None
    ) -> KnownCustomEmojiData:
        if not user and emoji.user:
            user = RefCell(copy.copy(emoji.user))

        # We ensure that all emojis that will get cached are guild based
        assert emoji.guild_id is not None

        return cls(
            id=emoji.id,
            name=emoji.name,
            is_animated=emoji.is_animated,
            guild_id=emoji.guild_id,
            is_colons_required=emoji.is_colons_required,
            is_managed=emoji.is_managed,
            is_available=emoji.is_available,
            user=user,
            # role_ids is a special case as it may be a mutable sequence so we want to ensure it's
            # immutable when cached.
            role_ids=tuple(emoji.role_ids),
        )

    @typing_extensions.override
    def build_entity(self, app: traits.RESTAware, /) -> emojis.KnownCustomEmoji:
        return emojis.KnownCustomEmoji(
            id=self.id,
            name=self.name,
            is_animated=self.is_animated,
            guild_id=self.guild_id,
            role_ids=self.role_ids,
            is_colons_required=self.is_colons_required,
            is_managed=self.is_managed,
            is_available=self.is_available,
            app=app,
            user=self.user.copy() if self.user else None,
        )


@attrs_extensions.with_copy
@attrs.define(kw_only=True, repr=False, weakref_slot=False)
class GuildStickerData(BaseData[stickers_.GuildSticker]):
    """A data model for storing sticker data in an in-memory cache."""

    id: snowflakes.Snowflake = attrs.field()
    name: str = attrs.field()
    description: str | None = attrs.field()
    tag: str = attrs.field()
    format_type: stickers_.StickerFormatType | int = attrs.field()
    is_available: bool = attrs.field()
    guild_id: snowflakes.Snowflake = attrs.field()
    user: RefCell[users_.User] | None = attrs.field()

    @classmethod
    @typing_extensions.override
    def build_from_entity(
        cls, sticker: stickers_.GuildSticker, /, *, user: RefCell[users_.User] | None = None
    ) -> GuildStickerData:
        if not user and sticker.user:
            user = RefCell(copy.copy(sticker.user))

        return cls(
            id=sticker.id,
            name=sticker.name,
            description=sticker.description,
            guild_id=sticker.guild_id,
            tag=sticker.tag,
            is_available=sticker.is_available,
            format_type=sticker.format_type,
            user=user,
        )

    @typing_extensions.override
    def build_entity(self, app: traits.RESTAware, /) -> stickers_.GuildSticker:
        return stickers_.GuildSticker(
            id=self.id,
            name=self.name,
            description=self.description,
            guild_id=self.guild_id,
            tag=self.tag,
            is_available=self.is_available,
            format_type=self.format_type,
            user=self.user.copy() if self.user else None,
        )


@attrs_extensions.with_copy
@attrs.define(kw_only=True, repr=False, weakref_slot=False)
class RichActivityData(BaseData[presences.RichActivity]):
    """A data model for storing rich activity data in an in-memory cache."""

    name: str = attrs.field()
    url: str | None = attrs.field()
    type: presences.ActivityType | int = attrs.field()
    created_at: datetime.datetime = attrs.field()
    timestamps: presences.ActivityTimestamps | None = attrs.field()
    application_id: snowflakes.Snowflake | None = attrs.field()
    details: str | None = attrs.field()
    state: str | None = attrs.field()
    emoji: RefCell[emojis.CustomEmoji] | str | None = attrs.field()
    party: presences.ActivityParty | None = attrs.field()
    assets: presences.ActivityAssets | None = attrs.field()
    secrets: presences.ActivitySecret | None = attrs.field()
    is_instance: bool | None = attrs.field()
    flags: presences.ActivityFlag | None = attrs.field()
    buttons: tuple[str, ...] = attrs.field()

    @classmethod
    @typing_extensions.override
    def build_from_entity(
        cls, activity: presences.RichActivity, /, *, emoji: RefCell[emojis.CustomEmoji] | str | None = None
    ) -> RichActivityData:
        if emoji:
            pass

        elif isinstance(activity.emoji, emojis.CustomEmoji):
            emoji = RefCell(copy.copy(activity.emoji))

        elif activity.emoji:
            emoji = activity.emoji.name

        timestamps = copy.copy(activity.timestamps) if activity.timestamps is not None else None
        party = copy.copy(activity.party) if activity.party is not None else None
        assets = copy.copy(activity.assets) if activity.assets is not None else None
        secrets = copy.copy(activity.secrets) if activity.secrets is not None else None
        return cls(
            name=activity.name,
            url=activity.url,
            type=activity.type,
            created_at=activity.created_at,
            application_id=activity.application_id,
            details=activity.details,
            state=activity.state,
            is_instance=activity.is_instance,
            flags=activity.flags,
            emoji=emoji,
            timestamps=timestamps,
            party=party,
            assets=assets,
            secrets=secrets,
            buttons=tuple(activity.buttons),
        )

    @typing_extensions.override
    def build_entity(self, _: traits.RESTAware, /) -> presences.RichActivity:
        emoji: emojis.Emoji | None = None
        if isinstance(self.emoji, RefCell):
            emoji = self.emoji.copy()

        elif self.emoji is not None:
            emoji = emojis.UnicodeEmoji(self.emoji)

        return presences.RichActivity(
            name=self.name,
            url=self.url,
            type=self.type,
            created_at=self.created_at,
            application_id=self.application_id,
            details=self.details,
            is_instance=self.is_instance,
            flags=self.flags,
            state=self.state,
            timestamps=copy.copy(self.timestamps) if self.timestamps is not None else None,
            party=copy.copy(self.party) if self.party is not None else None,
            assets=copy.copy(self.assets) if self.assets is not None else None,
            secrets=copy.copy(self.secrets) if self.secrets is not None else None,
            emoji=emoji,
            buttons=self.buttons,
        )


@attrs_extensions.with_copy
@attrs.define(kw_only=True, repr=False, weakref_slot=False)
class MemberPresenceData(BaseData[presences.MemberPresence]):
    """A data model for storing presence data in an in-memory cache."""

    user_id: snowflakes.Snowflake = attrs.field()
    guild_id: snowflakes.Snowflake = attrs.field()
    visible_status: presences.Status | str = attrs.field()
    activities: tuple[RichActivityData, ...] = attrs.field()
    client_status: presences.ClientStatus = attrs.field()

    @classmethod
    @typing_extensions.override
    def build_from_entity(cls, presence: presences.MemberPresence, /) -> MemberPresenceData:
        # role_ids and activities are special cases as may be mutable sequences, therefore we want to ensure they're
        # stored in immutable sequences (tuples). Plus activities need to be converted to Data objects.
        return cls(
            user_id=presence.user_id,
            guild_id=presence.guild_id,
            visible_status=presence.visible_status,
            activities=tuple(RichActivityData.build_from_entity(activity) for activity in presence.activities),
            client_status=copy.copy(presence.client_status),
        )

    @typing_extensions.override
    def build_entity(self, app: traits.RESTAware, /) -> presences.MemberPresence:
        return presences.MemberPresence(
            user_id=self.user_id,
            guild_id=self.guild_id,
            visible_status=self.visible_status,
            app=app,
            activities=[activity.build_entity(app) for activity in self.activities],
            client_status=copy.copy(self.client_status),
        )


def _copy_embed(embed: embeds_.Embed) -> embeds_.Embed:
    return embeds_.Embed.from_received_embed(
        title=embed.title,
        description=embed.description,
        url=embed.url,
        color=embed.color,
        timestamp=embed.timestamp,
        image=copy.copy(embed.image) if embed.image else None,
        thumbnail=copy.copy(embed.thumbnail) if embed.thumbnail else None,
        video=copy.copy(embed.video) if embed.video else None,
        author=copy.copy(embed.author) if embed.author else None,
        provider=copy.copy(embed.provider) if embed.provider else None,
        footer=copy.copy(embed.footer) if embed.footer else None,
        fields=[copy.copy(field) for field in embed.fields],
    )


@attrs_extensions.with_copy
@attrs.define(kw_only=True, repr=False, weakref_slot=False)
class MessageData(BaseData[messages.Message]):
    """A model for storing message data in an in-memory cache."""

    id: snowflakes.Snowflake = attrs.field()
    channel_id: snowflakes.Snowflake = attrs.field()
    guild_id: snowflakes.Snowflake | None = attrs.field()
    author: RefCell[users_.User] = attrs.field()
    member: RefCell[MemberData] | None = attrs.field()
    content: str | None = attrs.field()
    timestamp: datetime.datetime = attrs.field()
    edited_timestamp: datetime.datetime | None = attrs.field()
    is_tts: bool = attrs.field()
    user_mentions: undefined.UndefinedOr[typing.Mapping[snowflakes.Snowflake, RefCell[users_.User]]] = attrs.field()
    role_mention_ids: undefined.UndefinedOr[tuple[snowflakes.Snowflake, ...]] = attrs.field()
    channel_mentions: undefined.UndefinedOr[typing.Mapping[snowflakes.Snowflake, channels_.PartialChannel]] = (
        attrs.field()
    )
    mentions_everyone: undefined.UndefinedOr[bool] = attrs.field()
    attachments: tuple[messages.Attachment, ...] = attrs.field()
    embeds: tuple[embeds_.Embed, ...] = attrs.field()
    reactions: tuple[messages.Reaction, ...] = attrs.field()
    poll: polls_.Poll | None = attrs.field()
    is_pinned: bool = attrs.field()
    webhook_id: snowflakes.Snowflake | None = attrs.field()
    type: messages.MessageType | int = attrs.field()
    activity: messages.MessageActivity | None = attrs.field()
    application: messages.MessageApplication | None = attrs.field()
    message_reference: messages.MessageReference | None = attrs.field()
    flags: messages.MessageFlag = attrs.field()
    stickers: tuple[stickers_.PartialSticker, ...] = attrs.field()
    nonce: str | None = attrs.field()
    referenced_message: RefCell[MessageData] | None = attrs.field()
    message_snapshots: tuple[messages.MessageSnapshot, ...] = attrs.field()
    application_id: snowflakes.Snowflake | None = attrs.field()
    components: tuple[components_.TopLevelComponentTypesT, ...] = attrs.field()
    thread: channels_.GuildThreadChannel | None = attrs.field()
    interaction_metadata: base_interactions.PartialInteractionMetadata | None = attrs.field()

    @classmethod
    @typing_extensions.override
    def build_from_entity(
        cls,
        message: messages.Message,
        /,
        *,
        author: RefCell[users_.User] | None = None,
        member: RefCell[MemberData] | None = None,
        user_mentions: undefined.UndefinedOr[
            typing.Mapping[snowflakes.Snowflake, RefCell[users_.User]]
        ] = undefined.UNDEFINED,
        referenced_message: RefCell[MessageData] | None = None,
    ) -> MessageData:
        if not member and message.member:
            member = RefCell(MemberData.build_from_entity(message.member))

        if not user_mentions and message.user_mentions is not undefined.UNDEFINED:
            user_mentions = {user_id: RefCell(copy.copy(user)) for user_id, user in message.user_mentions.items()}

        channel_mentions: undefined.UndefinedOr[typing.Mapping[snowflakes.Snowflake, channels_.PartialChannel]] = (
            {channel_id: copy.copy(channel) for channel_id, channel in message.channel_mentions.items()}
            if message.channel_mentions is not undefined.UNDEFINED
            else undefined.UNDEFINED
        )
        role_mention_ids: undefined.UndefinedOr[tuple[snowflakes.Snowflake, ...]] = (
            tuple(message.role_mention_ids)
            if message.role_mention_ids is not undefined.UNDEFINED
            else undefined.UNDEFINED
        )

        return cls(
            id=message.id,
            channel_id=message.channel_id,
            guild_id=message.guild_id,
            author=author or RefCell(copy.copy(message.author)),
            member=member,
            content=message.content,
            timestamp=message.timestamp,
            edited_timestamp=message.edited_timestamp,
            is_tts=message.is_tts,
            user_mentions=user_mentions,
            channel_mentions=channel_mentions,
            role_mention_ids=role_mention_ids,
            mentions_everyone=message.mentions_everyone,
            attachments=tuple(map(copy.copy, message.attachments)),
            embeds=tuple(map(_copy_embed, message.embeds)),
            poll=message.poll,
            reactions=tuple(map(copy.copy, message.reactions)),
            is_pinned=message.is_pinned,
            webhook_id=message.webhook_id,
            type=message.type,
            activity=copy.copy(message.activity) if message.activity else None,
            application=copy.copy(message.application) if message.application else None,
            message_reference=copy.copy(message.message_reference) if message.message_reference else None,
            flags=copy.copy(message.flags),
            stickers=tuple(map(copy.copy, message.stickers)),
            nonce=message.nonce,
            referenced_message=referenced_message,
            message_snapshots=tuple(map(copy.copy, message.message_snapshots)),
            application_id=message.application_id,
            components=tuple(message.components),
            thread=message.thread,
            interaction_metadata=message.interaction_metadata,
        )

    @typing_extensions.override
    def build_entity(self, app: traits.RESTAware, /) -> messages.Message:
        channel_mentions: undefined.UndefinedOr[typing.Mapping[snowflakes.Snowflake, channels_.PartialChannel]] = (
            {channel_id: copy.copy(channel) for channel_id, channel in self.channel_mentions.items()}
            if self.channel_mentions is not undefined.UNDEFINED
            else undefined.UNDEFINED
        )
        user_mentions: undefined.UndefinedOr[typing.Mapping[snowflakes.Snowflake, users_.User]] = (
            {user_id: user.copy() for user_id, user in self.user_mentions.items()}
            if self.user_mentions is not undefined.UNDEFINED
            else undefined.UNDEFINED
        )

        return messages.Message(
            id=self.id,
            app=app,
            channel_id=self.channel_id,
            guild_id=self.guild_id,
            author=self.author.copy(),
            member=self.member.object.build_entity(app) if self.member else None,
            content=self.content,
            timestamp=self.timestamp,
            edited_timestamp=self.edited_timestamp,
            is_tts=self.is_tts,
            user_mentions=user_mentions,
            channel_mentions=channel_mentions,
            role_mention_ids=copy.copy(self.role_mention_ids),
            mentions_everyone=self.mentions_everyone,
            attachments=tuple(map(copy.copy, self.attachments)),
            embeds=tuple(map(_copy_embed, self.embeds)),
            poll=self.poll,
            reactions=tuple(map(copy.copy, self.reactions)),
            is_pinned=self.is_pinned,
            webhook_id=self.webhook_id,
            type=self.type,
            activity=copy.copy(self.activity) if self.activity else None,
            application=copy.copy(self.application) if self.application else None,
            message_reference=copy.copy(self.message_reference) if self.message_reference else None,
            flags=self.flags,
            stickers=tuple(map(copy.copy, self.stickers)),
            nonce=self.nonce,
            referenced_message=self.referenced_message.object.build_entity(app) if self.referenced_message else None,
            message_snapshots=self.message_snapshots,
            application_id=self.application_id,
            components=self.components,
            thread=self.thread,
            interaction_metadata=self.interaction_metadata,
        )

    def update(
        self,
        message: messages.PartialMessage,
        /,
        *,
        user_mentions: undefined.UndefinedOr[
            typing.Mapping[snowflakes.Snowflake, RefCell[users_.User]]
        ] = undefined.UNDEFINED,
    ) -> None:
        if message.content is not undefined.UNDEFINED:
            self.content = message.content

        if message.edited_timestamp is not undefined.UNDEFINED:
            self.edited_timestamp = message.edited_timestamp

        if message.is_pinned is not undefined.UNDEFINED:
            self.is_pinned = message.is_pinned

        if message.attachments is not undefined.UNDEFINED:
            self.attachments = tuple(map(copy.copy, message.attachments))

        if message.embeds is not undefined.UNDEFINED:
            self.embeds = tuple(map(_copy_embed, message.embeds))

        if message.components is not undefined.UNDEFINED:
            self.components = tuple(message.components)

        if user_mentions is not undefined.UNDEFINED:
            self.user_mentions = user_mentions
        elif message.user_mentions is not undefined.UNDEFINED:
            self.user_mentions = {user_id: RefCell(copy.copy(user)) for user_id, user in message.user_mentions.items()}

        if message.role_mention_ids is not undefined.UNDEFINED:
            self.role_mention_ids = tuple(message.role_mention_ids)

        if message.channel_mentions is not undefined.UNDEFINED:
            self.channel_mentions = {
                channel_id: copy.copy(channel) for channel_id, channel in message.channel_mentions.items()
            }

        if message.mentions_everyone is not undefined.UNDEFINED:
            self.mentions_everyone = message.mentions_everyone


@attrs_extensions.with_copy
@attrs.define(kw_only=True, repr=False, weakref_slot=False)
class VoiceStateData(BaseData[voices.VoiceState]):
    """A data model for storing voice state data in an in-memory cache."""

    channel_id: snowflakes.Snowflake | None = attrs.field()
    guild_id: snowflakes.Snowflake = attrs.field()
    user_id: snowflakes.Snowflake = attrs.field()
    is_guild_deafened: bool = attrs.field()
    is_guild_muted: bool = attrs.field()
    is_self_deafened: bool = attrs.field()
    is_self_muted: bool = attrs.field()
    is_streaming: bool = attrs.field()
    is_suppressed: bool = attrs.field()
    is_video_enabled: bool = attrs.field()
    member: RefCell[MemberData] | None = attrs.field()
    session_id: str = attrs.field()
    requested_to_speak_at: datetime.datetime | None = attrs.field()

    @typing_extensions.override
    def build_entity(self, app: traits.RESTAware, /) -> voices.VoiceState:
        member = self.member.object.build_entity(app) if self.member else None
        return voices.VoiceState(
            app=app,
            channel_id=self.channel_id,
            guild_id=self.guild_id,
            is_guild_deafened=self.is_guild_deafened,
            is_guild_muted=self.is_guild_muted,
            is_self_deafened=self.is_self_deafened,
            is_self_muted=self.is_self_muted,
            is_streaming=self.is_streaming,
            is_suppressed=self.is_suppressed,
            is_video_enabled=self.is_video_enabled,
            user_id=self.user_id,
            session_id=self.session_id,
            member=member,
            requested_to_speak_at=self.requested_to_speak_at,
        )

    @classmethod
    @typing_extensions.override
    def build_from_entity(
        cls, voice_state: voices.VoiceState, /, *, member: RefCell[MemberData] | None = None
    ) -> VoiceStateData:
        return cls(
            channel_id=voice_state.channel_id,
            guild_id=voice_state.guild_id,
            user_id=voice_state.user_id,
            is_self_deafened=voice_state.is_self_deafened,
            is_self_muted=voice_state.is_self_muted,
            is_guild_deafened=voice_state.is_guild_deafened,
            is_guild_muted=voice_state.is_guild_muted,
            is_streaming=voice_state.is_streaming,
            is_suppressed=voice_state.is_suppressed,
            is_video_enabled=voice_state.is_video_enabled,
            member=member,
            session_id=voice_state.session_id,
            requested_to_speak_at=voice_state.requested_to_speak_at,
        )


@attrs_extensions.with_copy
@attrs.define(repr=True, weakref_slot=True)
class Cell(typing.Generic[ValueT]):
    """Object used to store mutable references to a value in multiple places."""

    object: ValueT = attrs.field(repr=True)

    def copy(self) -> ValueT:
        """Get a copy of the contents of this cell.

        Returns
        -------
        ValueT
            The copied contents of this cell.
        """
        return copy.copy(self.object)


@attrs_extensions.with_copy
@attrs.define(repr=False, weakref_slot=False)
class RefCell(typing.Generic[ValueT]):
    """Object used to track mutable references to a value in multiple places.

    This is intended to enable reference counting for entities that are only kept
    alive by reference (e.g. the unknown emoji objects attached to presence
    activities and user objects) without the use of a "Data" object which lowers
    the time spent building these entities for the objects that reference them.
    """

    object: ValueT = attrs.field(repr=True)
    ref_count: int = attrs.field(default=0, kw_only=True)

    def copy(self) -> ValueT:
        """Get a copy of the contents of this cell.

        Returns
        -------
        ValueT
            The copied contents of this cell.
        """
        return copy.copy(self.object)


def unwrap_ref_cell(cell: RefCell[ValueT]) -> ValueT:
    """Unwrap a [`RefCell`][] instance to it's contents.

    Parameters
    ----------
    cell
        The reference cell instance to unwrap.

    Returns
    -------
    ValueT
        The reference cell's content.
    """
    return cell.copy()


def copy_guild_channel(channel: ChannelT) -> ChannelT:
    """Logic for handling the copying of guild channel objects.

    This exists account for the permission overwrite objects attached to guild
    channel objects which need to be copied themselves.
    """
    channel = copy.copy(channel)
    channel.permission_overwrites = {
        sf: copy.copy(overwrite) for sf, overwrite in channel.permission_overwrites.items()
    }
    return channel


class Cache3DMappingView(CacheMappingView[snowflakes.Snowflake, cache.CacheView[KeyT, ValueT]]):
    """A special case of the Mapping View which avoids copying the immutable values contained within it."""

    __slots__: typing.Sequence[str] = ()

    @staticmethod
    @typing_extensions.override
    def _copy(value: cache.CacheView[KeyT, ValueT]) -> cache.CacheView[KeyT, ValueT]:
        return value



# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import mock
import pytest

from hikari import emojis
from hikari import guilds
from hikari.events import reaction_events
from tests.hikari import hikari_test_helpers


class TestReactionAddEvent:
    def test_is_for_emoji_when_custom_emoji_matches(self):
        event = hikari_test_helpers.mock_class_namespace(reaction_events.ReactionAddEvent, emoji_id=333333)()

        assert event.is_for_emoji(emojis.CustomEmoji(id=333333, name=None, is_animated=True))

    def test_is_for_emoji_when_unicode_emoji_matches(self):
        event = hikari_test_helpers.mock_class_namespace(reaction_events.ReactionAddEvent, emoji_name="🌲")()

        assert event.is_for_emoji(emojis.UnicodeEmoji("🌲"))

    @pytest.mark.parametrize(
        ("emoji_id", "emoji_name", "emoji"),
        [
            (None, "hi", emojis.CustomEmoji(name=None, id=54123, is_animated=False)),
            (123321, None, emojis.UnicodeEmoji("no u")),
        ],
    )
    def test_is_for_emoji_when_wrong_emoji_type(self, emoji_id, emoji_name, emoji):
        event = hikari_test_helpers.mock_class_namespace(
            reaction_events.ReactionAddEvent, emoji_id=emoji_id, emoji_name=emoji_name
        )()

        assert event.is_for_emoji(emoji) is False

    @pytest.mark.parametrize(
        ("emoji_id", "emoji_name", "emoji"),
        [
            (None, "hi", emojis.UnicodeEmoji("bye")),
            (123321, None, emojis.CustomEmoji(id=123312123, name=None, is_animated=False)),
        ],
    )
    def test_is_for_emoji_when_emoji_miss_match(self, emoji_id, emoji_name, emoji):
        event = hikari_test_helpers.mock_class_namespace(
            reaction_events.ReactionAddEvent, emoji_id=emoji_id, emoji_name=emoji_name
        )()

        assert event.is_for_emoji(emoji) is False


class TestReactionDeleteEvent:
    def test_is_for_emoji_when_custom_emoji_matches(self):
        event = hikari_test_helpers.mock_class_namespace(reaction_events.ReactionDeleteEvent, emoji_id=333)()

        assert event.is_for_emoji(emojis.CustomEmoji(id=333, name=None, is_animated=True))

    def test_is_for_emoji_when_unicode_emoji_matches(self):
        event = hikari_test_helpers.mock_class_namespace(reaction_events.ReactionDeleteEvent, emoji_name="e")()

        assert event.is_for_emoji(emojis.UnicodeEmoji("e"))

    @pytest.mark.parametrize(
        ("emoji_id", "emoji_name", "emoji"),
        [
            (None, "hasdi", emojis.CustomEmoji(name=None, id=3123, is_animated=False)),
            (534123, None, emojis.UnicodeEmoji("nodfgdu")),
        ],
    )
    def test_is_for_emoji_when_wrong_emoji_type(self, emoji_id, emoji_name, emoji):
        event = hikari_test_helpers.mock_class_namespace(
            reaction_events.ReactionDeleteEvent, emoji_id=emoji_id, emoji_name=emoji_name
        )()

        assert event.is_for_emoji(emoji) is False

    @pytest.mark.parametrize(
        ("emoji_id", "emoji_name", "emoji"),
        [
            (None, "hfdasi", emojis.UnicodeEmoji("bgye")),
            (54123, None, emojis.CustomEmoji(id=34123, name=None, is_animated=False)),
        ],
    )
    def test_is_for_emoji_when_emoji_miss_match(self, emoji_id, emoji_name, emoji):
        event = hikari_test_helpers.mock_class_namespace(
            reaction_events.ReactionDeleteEvent, emoji_id=emoji_id, emoji_name=emoji_name
        )()

        assert event.is_for_emoji(emoji) is False


class TestReactionDeleteEmojiEvent:
    def test_is_for_emoji_when_custom_emoji_matches(self):
        event = hikari_test_helpers.mock_class_namespace(reaction_events.ReactionDeleteEmojiEvent, emoji_id=332223333)()

        assert event.is_for_emoji(emojis.CustomEmoji(id=332223333, name=None, is_animated=True))

    def test_is_for_emoji_when_unicode_emoji_matches(self):
        event = hikari_test_helpers.mock_class_namespace(reaction_events.ReactionDeleteEmojiEvent, emoji_name="🌲e")()

        assert event.is_for_emoji(emojis.UnicodeEmoji("🌲e"))

    @pytest.mark.parametrize(
        ("emoji_id", "emoji_name", "emoji"),
        [
            (None, "heeei", emojis.CustomEmoji(name=None, id=541123, is_animated=False)),
            (1233211, None, emojis.UnicodeEmoji("no eeeu")),
        ],
    )
    def test_is_for_emoji_when_wrong_emoji_type(self, emoji_id, emoji_name, emoji):
        event = hikari_test_helpers.mock_class_namespace(
            reaction_events.ReactionDeleteEmojiEvent, emoji_id=emoji_id, emoji_name=emoji_name
        )()

        assert event.is_for_emoji(emoji) is False

    @pytest.mark.parametrize(
        ("emoji_id", "emoji_name", "emoji"),
        [
            (None, "dsahi", emojis.UnicodeEmoji("bye321")),
            (12331231, None, emojis.CustomEmoji(id=121233312123, name=None, is_animated=False)),
        ],
    )
    def test_is_for_emoji_when_emoji_miss_match(self, emoji_id, emoji_name, emoji):
        event = hikari_test_helpers.mock_class_namespace(
            reaction_events.ReactionDeleteEmojiEvent, emoji_id=emoji_id, emoji_name=emoji_name
        )()

        assert event.is_for_emoji(emoji) is False


class TestGuildReactionAddEvent:
    @pytest.fixture
    def event(self):
        return reaction_events.GuildReactionAddEvent(
            shard=object(),
            member=mock.MagicMock(guilds.Member),
            channel_id=123,
            message_id=456,
            emoji_name="👌",
            emoji_id=None,
            is_animated=False,
        )

    def test_app_property(self, event):
        assert event.app is event.member.app

    def test_guild_id_property(self, event):
        event.member.guild_id = 123
        assert event.guild_id == 123

    def test_user_id_property(self, event):
        event.member.user.id = 123
        assert event.user_id == 123



# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import mock
import pytest

from hikari import channels
from hikari import snowflakes
from hikari.events import channel_events
from tests.hikari import hikari_test_helpers


class TestGuildChannelEvent:
    @pytest.fixture
    def event(self):
        cls = hikari_test_helpers.mock_class_namespace(
            channel_events.GuildChannelEvent,
            guild_id=mock.PropertyMock(return_value=snowflakes.Snowflake(929292929)),
            channel_id=mock.PropertyMock(return_value=snowflakes.Snowflake(432432432)),
        )
        return cls()

    def test_get_guild_when_available(self, event):
        result = event.get_guild()

        assert result is event.app.cache.get_available_guild.return_value
        event.app.cache.get_available_guild.assert_called_once_with(929292929)
        event.app.cache.get_unavailable_guild.assert_not_called()

    def test_get_guild_when_unavailable(self, event):
        event.app.cache.get_available_guild.return_value = None
        result = event.get_guild()

        assert result is event.app.cache.get_unavailable_guild.return_value
        event.app.cache.get_available_guild.assert_called_once_with(929292929)
        event.app.cache.get_unavailable_guild.assert_called_once_with(929292929)

    def test_get_guild_without_cache(self):
        event = hikari_test_helpers.mock_class_namespace(channel_events.GuildChannelEvent, app=None)()

        assert event.get_guild() is None

    @pytest.mark.asyncio
    async def test_fetch_guild(self, event):
        event.app.rest.fetch_guild = mock.AsyncMock()
        result = await event.fetch_guild()

        assert result is event.app.rest.fetch_guild.return_value
        event.app.rest.fetch_guild.assert_awaited_once_with(929292929)

    def test_get_channel(self, event):
        result = event.get_channel()

        assert result is event.app.cache.get_guild_channel.return_value
        event.app.cache.get_guild_channel.assert_called_once_with(432432432)

    def test_get_channel_without_cache(self):
        event = hikari_test_helpers.mock_class_namespace(channel_events.GuildChannelEvent, app=None)()

        assert event.get_channel() is None

    @pytest.mark.asyncio
    async def test_fetch_channel(self, event):
        event.app.rest.fetch_channel = mock.AsyncMock(return_value=mock.MagicMock(spec=channels.GuildChannel))
        result = await event.fetch_channel()

        assert result is event.app.rest.fetch_channel.return_value
        event.app.rest.fetch_channel.assert_awaited_once_with(432432432)


class TestGuildChannelCreateEvent:
    @pytest.fixture
    def event(self):
        return channel_events.GuildChannelCreateEvent(channel=mock.Mock(), shard=None)

    def test_app_property(self, event):
        assert event.app is event.channel.app

    def test_channel_id_property(self, event):
        event.channel.id = 123
        assert event.channel_id == 123

    def test_guild_id_property(self, event):
        event.channel.guild_id = 123
        assert event.guild_id == 123


class TestGuildChannelUpdateEvent:
    @pytest.fixture
    def event(self):
        return channel_events.GuildChannelUpdateEvent(channel=mock.Mock(), old_channel=mock.Mock(), shard=None)

    def test_app_property(self, event):
        assert event.app is event.channel.app

    def test_channel_id_property(self, event):
        event.channel.id = 123
        assert event.channel_id == 123

    def test_guild_id_property(self, event):
        event.channel.guild_id = 123
        assert event.guild_id == 123

    def test_old_channel_id_property(self, event):
        event.old_channel.id = 123
        assert event.old_channel.id == 123


class TestGuildChannelDeleteEvent:
    @pytest.fixture
    def event(self):
        return channel_events.GuildChannelDeleteEvent(channel=mock.Mock(), shard=None)

    def test_app_property(self, event):
        assert event.app is event.channel.app

    def test_channel_id_property(self, event):
        event.channel.id = 123
        assert event.channel_id == 123

    def test_guild_id_property(self, event):
        event.channel.guild_id = 123
        assert event.guild_id == 123


class TestGuildPinsUpdateEvent:
    @pytest.fixture
    def event(self):
        return channel_events.GuildPinsUpdateEvent(
            app=mock.Mock(), shard=None, channel_id=12343, guild_id=None, last_pin_timestamp=None
        )

    @pytest.mark.parametrize("result", [mock.Mock(spec=channels.GuildTextChannel), None])
    def test_get_channel(self, event, result):
        event.app.cache.get_guild_channel.return_value = result

        result = event.get_channel()

        assert result is event.app.cache.get_guild_channel.return_value
        event.app.cache.get_guild_channel.assert_called_once_with(event.channel_id)


@pytest.mark.asyncio
class TestInviteEvent:
    @pytest.fixture
    def event(self):
        return hikari_test_helpers.mock_class_namespace(
            channel_events.InviteEvent, slots_=False, code=mock.PropertyMock(return_value="Jx4cNGG")
        )()

    async def test_fetch_invite(self, event):
        event.app.rest.fetch_invite = mock.AsyncMock()

        await event.fetch_invite()

        event.app.rest.fetch_invite.assert_awaited_once_with("Jx4cNGG")


class TestInviteCreateEvent:
    @pytest.fixture
    def event(self):
        return channel_events.InviteCreateEvent(shard=None, invite=mock.Mock())

    def test_app_property(self, event):
        assert event.app is event.invite.app

    @pytest.mark.asyncio
    async def test_channel_id_property(self, event):
        event.invite.channel_id = 123
        assert event.channel_id == 123

    @pytest.mark.asyncio
    async def test_guild_id_property(self, event):
        event.invite.guild_id = 123
        assert event.guild_id == 123

    @pytest.mark.asyncio
    async def test_code_property(self, event):
        event.invite.code = "Jx4cNGG"
        assert event.code == "Jx4cNGG"


@pytest.mark.asyncio
class TestWebhookUpdateEvent:
    @pytest.fixture
    def event(self):
        return channel_events.WebhookUpdateEvent(app=mock.AsyncMock(), shard=mock.Mock(), channel_id=123, guild_id=456)

    async def test_fetch_channel_webhooks(self, event):
        await event.fetch_channel_webhooks()

        event.app.rest.fetch_channel_webhooks.assert_awaited_once_with(123)

    async def test_fetch_guild_webhooks(self, event):
        await event.fetch_guild_webhooks()

        event.app.rest.fetch_guild_webhooks.assert_awaited_once_with(456)


class TestGuildThreadEvent:
    @pytest.mark.asyncio
    async def test_fetch_channel(self):
        mock_app = mock.AsyncMock()
        mock_app.rest.fetch_channel.return_value = mock.Mock(channels.GuildThreadChannel)
        event = hikari_test_helpers.mock_class_namespace(
            channel_events.GuildThreadEvent, app=mock_app, thread_id=123321
        )()

        result = await event.fetch_channel()

        assert result is mock_app.rest.fetch_channel.return_value
        mock_app.rest.fetch_channel.assert_awaited_once_with(123321)


class TestGuildThreadAccessEvent:
    @pytest.fixture
    def event(self) -> channel_events.GuildThreadAccessEvent:
        return channel_events.GuildThreadAccessEvent(shard=mock.Mock(), thread=mock.Mock())

    def test_app_property(self, event: channel_events.GuildThreadAccessEvent):
        assert event.app is event.thread.app

    def test_guild_id(self, event: channel_events.GuildThreadAccessEvent):
        assert event.guild_id is event.thread.guild_id

    def test_thread_id_property(self, event: channel_events.GuildThreadAccessEvent):
        assert event.thread_id is event.thread.id


class TestGuildThreadCreateEvent:
    @pytest.fixture
    def event(self) -> channel_events.GuildThreadCreateEvent:
        return channel_events.GuildThreadCreateEvent(shard=mock.Mock(), thread=mock.Mock())

    def test_app_property(self, event: channel_events.GuildThreadCreateEvent):
        assert event.app is event.thread.app

    def test_guild_id(self, event: channel_events.GuildThreadCreateEvent):
        assert event.guild_id is event.thread.guild_id

    def test_thread_id_property(self, event: channel_events.GuildThreadCreateEvent):
        assert event.thread_id is event.thread.id


class TestGuildThreadUpdateEvent:
    @pytest.fixture
    def event(self) -> channel_events.GuildThreadUpdateEvent:
        return channel_events.GuildThreadUpdateEvent(shard=mock.Mock(), thread=mock.Mock(), old_thread=mock.Mock())

    def test_app_property(self, event: channel_events.GuildThreadUpdateEvent):
        assert event.app is event.thread.app

    def test_guild_id(self, event: channel_events.GuildThreadUpdateEvent):
        assert event.guild_id is event.thread.guild_id

    def test_thread_id_property(self, event: channel_events.GuildThreadUpdateEvent):
        assert event.thread_id is event.thread.id



# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import asyncio
import contextlib
import gc
import sys
import typing
import warnings
import weakref

import mock
import pytest

from hikari import errors
from hikari import intents
from hikari import iterators
from hikari.api import config
from hikari.events import base_events
from hikari.events import member_events
from hikari.events import shard_events
from hikari.impl import event_manager_base
from hikari.internal import reflect
from tests.hikari import hikari_test_helpers


class TestGenerateWeakListener:
    @pytest.mark.asyncio
    async def test__generate_weak_listener_when_method_is_None(self):
        def test():
            return None

        call_weak_method = event_manager_base._generate_weak_listener(test)

        with pytest.raises(
            TypeError,
            match=r"dead weak referenced subscriber method cannot be executed, try actually closing your event streamers",
        ):
            await call_weak_method(None)

    @pytest.mark.asyncio
    async def test__generate_weak_listener(self):
        mock_listener = mock.AsyncMock()
        mock_event = object()

        def test():
            return mock_listener

        call_weak_method = event_manager_base._generate_weak_listener(test)

        await call_weak_method(mock_event)

        mock_listener.assert_awaited_once_with(mock_event)


@pytest.fixture
def mock_app():
    return mock.Mock()


class TestEventStream:
    def test___enter___and___exit__(self):
        stub_stream = hikari_test_helpers.mock_class_namespace(
            event_manager_base.EventStream, open=mock.Mock(), close=mock.Mock()
        )(mock_app, base_events.Event, timeout=None)

        with stub_stream:
            stub_stream.open.assert_called_once_with()
            stub_stream.close.assert_not_called()

        stub_stream.open.assert_called_once_with()
        stub_stream.close.assert_called_once_with()

    @pytest.mark.asyncio
    async def test__listener_when_filter_returns_false(self, mock_app):
        stream = event_manager_base.EventStream(mock_app, base_events.Event, timeout=None)
        stream.filter(lambda _: False)
        mock_event = object()

        assert await stream._listener(mock_event) is None
        assert not stream._queue

    @hikari_test_helpers.timeout()
    @pytest.mark.asyncio
    async def test__listener_when_filter_passes_and_queue_full(self, mock_app):
        stream = event_manager_base.EventStream(mock_app, base_events.Event, timeout=None, limit=2)
        stream._queue.append(object())
        stream._queue.append(object())
        stream.filter(lambda _: True)
        mock_event = object()

        with stream:
            assert await stream._listener(mock_event) is None
            assert await stream.next() is not mock_event
            assert await stream.next() is not mock_event
            assert not stream._queue

    @hikari_test_helpers.timeout()
    @pytest.mark.asyncio
    async def test__listener_when_filter_passes_and_queue_not_full(self, mock_app):
        stream = event_manager_base.EventStream(mock_app, base_events.Event, timeout=None, limit=None)
        stream._queue.append(object())
        stream._queue.append(object())
        stream.filter(lambda _: True)
        mock_event = object()

        with stream:
            assert await stream._listener(mock_event) is None
            assert await stream.next() is not mock_event
            assert await stream.next() is not mock_event
            assert await stream.next() is mock_event

    @pytest.mark.asyncio
    @hikari_test_helpers.timeout()
    async def test___anext___when_stream_closed(self):
        streamer = event_manager_base.EventStream(mock.Mock(), event_type=base_events.Event, timeout=float("inf"))

        # ruff gets annoyed if we use "with" here so here's a hacky alternative
        with pytest.raises(TypeError):
            await streamer.__anext__()

    @pytest.mark.asyncio
    @hikari_test_helpers.timeout()
    async def test___anext___times_out(self):
        streamer = event_manager_base.EventStream(mock.Mock(), event_type=base_events.Event, timeout=0.001)

        with streamer:
            with pytest.raises(LookupError):
                await streamer.next()

    @pytest.mark.asyncio
    @hikari_test_helpers.timeout()
    async def test___anext___waits_for_next_event(self):
        mock_event = object()
        streamer = event_manager_base.EventStream(mock.Mock(), event_type=base_events.Event, timeout=None)

        async def quickly_run_task(task):
            try:
                await asyncio.wait_for(asyncio.shield(task), timeout=0.01)
            except asyncio.TimeoutError:
                pass

        with streamer:
            next_task = asyncio.create_task(streamer.next())
            await quickly_run_task(next_task)
            assert not next_task.done()
            await streamer._listener(mock_event)
            await quickly_run_task(next_task)
            assert next_task.done()
            assert next_task.result() is mock_event

    @pytest.mark.asyncio
    @hikari_test_helpers.timeout()
    async def test___anext__(self):
        mock_event = object()
        streamer = event_manager_base.EventStream(
            event_manager=mock.Mock(),
            event_type=base_events.Event,
            timeout=hikari_test_helpers.REASONABLE_QUICK_RESPONSE_TIME,
        )
        streamer._queue.append(mock_event)

        with streamer:
            assert await streamer.next() is mock_event

    @pytest.mark.asyncio
    async def test___await__(self):
        mock_event_0 = object()
        mock_event_1 = object()
        mock_event_2 = object()
        streamer = hikari_test_helpers.mock_class_namespace(
            event_manager_base.EventStream,
            close=mock.Mock(),
            open=mock.Mock(),
            init_=False,
            _active=False,
            __anext__=mock.AsyncMock(side_effect=[mock_event_0, mock_event_1, mock_event_2]),
        )()

        assert await streamer == [mock_event_0, mock_event_1, mock_event_2]
        streamer.open.assert_called_once_with()
        streamer.close.assert_called_once_with()

    def test___del___for_active_stream(self):
        mock_coroutine = object()
        close_method = mock.Mock(return_value=mock_coroutine)
        streamer = hikari_test_helpers.mock_class_namespace(
            event_manager_base.EventStream, close=close_method, init_=False
        )()
        streamer._event_type = base_events.Event
        streamer._active = True

        with mock.patch.object(event_manager_base, "_LOGGER") as logger:
            del streamer
            gc.collect()  # Force a GC collection

        logger.warning.assert_called_once_with("active %r streamer fell out of scope before being closed", "Event")
        close_method.assert_called_once_with()

    def test___del___for_inactive_stream(self):
        close_method = mock.Mock()
        streamer = hikari_test_helpers.mock_class_namespace(
            event_manager_base.EventStream, close=close_method, init_=False
        )()
        streamer._event_type = base_events.Event
        streamer._active = False

        del streamer
        close_method.assert_not_called()

    def test_close_for_inactive_stream(self, mock_app):
        stream = event_manager_base.EventStream(mock_app, base_events.Event, timeout=None, limit=None)
        stream.close()
        mock_app.event_manager.unsubscribe.assert_not_called()

    def test_close_for_active_stream(self):
        mock_registered_listener = object()
        mock_manager = mock.Mock()
        stream = hikari_test_helpers.mock_class_namespace(event_manager_base.EventStream)(
            event_manager=mock_manager, event_type=base_events.Event, timeout=float("inf")
        )

        stream.open()
        stream._registered_listener = mock_registered_listener
        stream.close()
        mock_manager.unsubscribe.assert_called_once_with(base_events.Event, mock_registered_listener)
        assert stream._active is False
        assert stream._registered_listener is None

    def test_close_for_active_stream_handles_value_error(self):
        mock_registered_listener = object()
        mock_manager = mock.Mock()
        mock_manager.unsubscribe.side_effect = ValueError
        stream = hikari_test_helpers.mock_class_namespace(event_manager_base.EventStream)(
            event_manager=mock_manager, event_type=base_events.Event, timeout=float("inf")
        )

        stream.open()
        stream._registered_listener = mock_registered_listener
        stream.close()
        mock_manager.unsubscribe.assert_called_once_with(base_events.Event, mock_registered_listener)
        assert stream._active is False
        assert stream._registered_listener is None

    @pytest.mark.asyncio
    async def test_filter(self):
        stream = hikari_test_helpers.mock_class_namespace(event_manager_base.EventStream)(
            event_manager=mock.Mock(), event_type=base_events.Event, timeout=0.001
        )
        stream._filters = iterators.All(())
        first_pass = mock.Mock(attr=True)
        second_pass = mock.Mock(attr=True)
        first_fails = mock.Mock(attr=True)
        second_fail = mock.Mock(attr=False)

        def predicate(obj):
            return obj in (first_pass, second_pass)

        stream.filter(predicate, attr=True)

        await stream._listener(first_pass)
        await stream._listener(first_fails)
        await stream._listener(second_pass)
        await stream._listener(second_fail)

        assert await stream == [first_pass, second_pass]

    @pytest.mark.asyncio
    async def test_filter_handles_calls_while_active(self):
        stream = hikari_test_helpers.mock_class_namespace(event_manager_base.EventStream)(
            event_manager=mock.Mock(), event_type=base_events.Event, timeout=0.001
        )
        stream._filters = iterators.All(())
        first_pass = mock.Mock(attr=True)
        second_pass = mock.Mock(attr=True)
        first_fails = mock.Mock(attr=True)
        second_fail = mock.Mock(attr=False)
        await stream._listener(first_pass)
        await stream._listener(first_fails)
        await stream._listener(second_pass)
        await stream._listener(second_fail)

        def predicate(obj):
            return obj in (first_pass, second_pass)

        with stream:
            stream.filter(predicate, attr=True)

            assert await stream == [first_pass, second_pass]

    def test_open_for_inactive_stream(self):
        mock_listener = object()
        mock_manager = mock.Mock()
        stream = hikari_test_helpers.mock_class_namespace(event_manager_base.EventStream)(
            event_manager=mock_manager, event_type=base_events.Event, timeout=float("inf")
        )

        stream._active = True
        stream._registered_listener = mock_listener

        with mock.patch.object(event_manager_base, "_generate_weak_listener"):
            with mock.patch.object(weakref, "WeakMethod"):
                stream.open()

                weakref.WeakMethod.assert_not_called()
            event_manager_base._generate_weak_listener.assert_not_called()

        mock_manager.subscribe.assert_not_called()
        assert stream._active is True
        assert stream._registered_listener is mock_listener

        # Ensure we don't get a warning or error on del
        stream._active = False

    def test_open_for_active_stream(self):
        mock_manager = mock.Mock()
        stream = hikari_test_helpers.mock_class_namespace(event_manager_base.EventStream)(
            event_manager=mock_manager, event_type=base_events.Event, timeout=float("inf")
        )
        stream._active = False
        mock_listener = object()
        mock_listener_ref = object()

        with mock.patch.object(event_manager_base, "_generate_weak_listener", return_value=mock_listener):
            with mock.patch.object(weakref, "WeakMethod", return_value=mock_listener_ref):
                stream.open()

                weakref.WeakMethod.assert_called_once_with(stream._listener)
            event_manager_base._generate_weak_listener.assert_called_once_with(mock_listener_ref)

        mock_manager.subscribe.assert_called_once_with(base_events.Event, mock_listener)
        assert stream._active is True
        assert stream._registered_listener is mock_listener

        # Ensure we don't get a warning or error on del
        stream._active = False


class TestConsumer:
    @pytest.mark.parametrize(
        ("is_caching", "listener_group_count", "waiter_group_count", "expected_result"),
        [(True, -10000, -10000, True), (False, 0, 1, True), (False, 1, 0, True), (False, 0, 0, False)],
    )
    def test_is_enabled(self, is_caching, listener_group_count, waiter_group_count, expected_result):
        consumer = event_manager_base._Consumer(object(), 123, is_caching)
        consumer.listener_group_count = listener_group_count
        consumer.waiter_group_count = waiter_group_count

        assert consumer.is_enabled is expected_result


class TestEventManagerBase:
    @pytest.fixture
    def event_manager(self):
        class EventManagerBaseImpl(event_manager_base.EventManagerBase):
            on_existing_event = None

        return EventManagerBaseImpl(mock.Mock(), mock.Mock())

    def test___init___loads_consumers(self):
        class StubManager(event_manager_base.EventManagerBase):
            @event_manager_base.filtered(shard_events.ShardEvent, config.CacheComponents.MEMBERS)
            async def on_foo(self, event):
                raise NotImplementedError

            @event_manager_base.filtered((shard_events.ShardStateEvent, shard_events.ShardPayloadEvent))
            async def on_bar(self, event):
                raise NotImplementedError

            @event_manager_base.filtered(shard_events.MemberChunkEvent, config.CacheComponents.MESSAGES)
            async def on_bat(self, event):
                raise NotImplementedError

            async def on_not_decorated(self, event):
                raise NotImplementedError

            async def not_a_listener(self):
                raise NotImplementedError

        manager = StubManager(
            mock.Mock(), 0, cache_components=config.CacheComponents.MEMBERS | config.CacheComponents.GUILD_CHANNELS
        )
        assert manager._consumers == {
            "foo": event_manager_base._Consumer(manager.on_foo, 9, True),
            "bar": event_manager_base._Consumer(manager.on_bar, 105, False),
            "bat": event_manager_base._Consumer(manager.on_bat, 65545, False),
            "not_decorated": event_manager_base._Consumer(manager.on_not_decorated, -1, True),
        }

    def test___init___loads_consumers_when_cacheless(self):
        class StubManager(event_manager_base.EventManagerBase):
            @event_manager_base.filtered(shard_events.ShardEvent, config.CacheComponents.MEMBERS)
            async def on_foo(self, event):
                raise NotImplementedError

            @event_manager_base.filtered((shard_events.ShardStateEvent, shard_events.ShardPayloadEvent))
            async def on_bar(self, event):
                raise NotImplementedError

            @event_manager_base.filtered(shard_events.MemberChunkEvent, config.CacheComponents.MESSAGES)
            async def on_bat(self, event):
                raise NotImplementedError

            async def on_not_decorated(self, event):
                raise NotImplementedError

            async def not_a_listener(self):
                raise NotImplementedError

        manager = StubManager(mock.Mock(), 0, cache_components=config.CacheComponents.NONE)
        assert manager._consumers == {
            "foo": event_manager_base._Consumer(manager.on_foo, 9, False),
            "bar": event_manager_base._Consumer(manager.on_bar, 105, False),
            "bat": event_manager_base._Consumer(manager.on_bat, 65545, False),
            "not_decorated": event_manager_base._Consumer(manager.on_not_decorated, -1, False),
        }

    def test__increment_listener_group_count(self, event_manager):
        on_foo_consumer = event_manager_base._Consumer(None, 9, False)
        on_bar_consumer = event_manager_base._Consumer(None, 105, False)
        on_bat_consumer = event_manager_base._Consumer(None, 1, False)
        event_manager._consumers = {"foo": on_foo_consumer, "bar": on_bar_consumer, "bat": on_bat_consumer}

        event_manager._increment_listener_group_count(shard_events.ShardEvent, 1)

        assert on_foo_consumer.listener_group_count == 1
        assert on_bar_consumer.listener_group_count == 1
        assert on_bat_consumer.listener_group_count == 0

    def test__increment_waiter_group_count(self, event_manager):
        on_foo_consumer = event_manager_base._Consumer(None, 9, False)
        on_bar_consumer = event_manager_base._Consumer(None, 105, False)
        on_bat_consumer = event_manager_base._Consumer(None, 1, False)
        event_manager._consumers = {"foo": on_foo_consumer, "bar": on_bar_consumer, "bat": on_bat_consumer}

        event_manager._increment_waiter_group_count(shard_events.ShardEvent, 1)

        assert on_foo_consumer.waiter_group_count == 1
        assert on_bar_consumer.waiter_group_count == 1
        assert on_bat_consumer.waiter_group_count == 0

    def test__enabled_for_event_when_listener_registered(self, event_manager):
        event_manager._listeners = {shard_events.ShardStateEvent: [], shard_events.MemberChunkEvent: []}
        event_manager._waiters = {}

        assert event_manager._enabled_for_event(shard_events.ShardStateEvent) is True

    def test__enabled_for_event_when_waiter_registered(self, event_manager):
        event_manager._listeners = {}
        event_manager._waiters = {shard_events.ShardStateEvent: [], shard_events.MemberChunkEvent: []}

        assert event_manager._enabled_for_event(shard_events.ShardStateEvent) is True

    def test__enabled_for_event_when_not_registered(self, event_manager):
        event_manager._listeners = {shard_events.ShardPayloadEvent: [], shard_events.MemberChunkEvent: []}
        event_manager._waiters = {shard_events.ShardPayloadEvent: [], shard_events.MemberChunkEvent: []}

        assert event_manager._enabled_for_event(shard_events.ShardStateEvent) is False

    @pytest.mark.asyncio
    async def test_consume_raw_event_when_KeyError(self, event_manager):
        event_manager._enabled_for_event = mock.Mock(return_value=True)
        mock_payload = {"id": "3123123123"}
        mock_shard = mock.Mock(id=123)
        event_manager.dispatch = mock.Mock()

        with pytest.raises(LookupError):
            event_manager.consume_raw_event("UNEXISTING_EVENT", mock_shard, mock_payload)

        event_manager.dispatch.assert_called_once_with(
            event_manager._event_factory.deserialize_shard_payload_event.return_value
        )
        event_manager._event_factory.deserialize_shard_payload_event.assert_called_once_with(
            mock_shard, mock_payload, name="UNEXISTING_EVENT"
        )
        event_manager._enabled_for_event.assert_called_once_with(shard_events.ShardPayloadEvent)

    @pytest.mark.asyncio
    async def test_consume_raw_event_when_found(self, event_manager):
        event_manager._enabled_for_event = mock.Mock(return_value=True)
        event_manager.dispatch = mock.Mock()
        on_existing_event = mock.Mock(is_enabled=True, callback=mock.Mock(__name__="testing"))
        event_manager._consumers = {"existing_event": on_existing_event}
        shard = object()
        payload = {"berp": "baz"}

        event_manager.consume_raw_event("EXISTING_EVENT", shard, payload)

        on_existing_event.callback.assert_called_once_with(shard, payload)
        event_manager.dispatch.assert_called_once_with(
            event_manager._event_factory.deserialize_shard_payload_event.return_value
        )
        event_manager._event_factory.deserialize_shard_payload_event.assert_called_once_with(
            shard, payload, name="EXISTING_EVENT"
        )
        event_manager._enabled_for_event.assert_called_once_with(shard_events.ShardPayloadEvent)

    @pytest.mark.asyncio
    async def test_consume_raw_event_skips_consumer_callback_when_not_enabled(self, event_manager):
        event_manager._enabled_for_event = mock.Mock(return_value=True)
        event_manager.dispatch = mock.Mock()
        on_existing_event = mock.Mock(is_enabled=False, callback=mock.Mock(__name__="testing"))
        event_manager._consumers = {"existing_event": on_existing_event}
        shard = object()
        payload = {"berp": "baz"}

        event_manager.consume_raw_event("EXISTING_EVENT", shard, payload)

        on_existing_event.callback.assert_not_called()
        event_manager.dispatch.assert_called_once_with(
            event_manager._event_factory.deserialize_shard_payload_event.return_value
        )
        event_manager._event_factory.deserialize_shard_payload_event.assert_called_once_with(
            shard, payload, name="EXISTING_EVENT"
        )
        event_manager._enabled_for_event.assert_called_once_with(shard_events.ShardPayloadEvent)

    @pytest.mark.asyncio
    async def test_consume_raw_event_skips_raw_dispatch_when_not_enabled(self, event_manager):
        event_manager._enabled_for_event = mock.Mock(return_value=False)
        event_manager.dispatch = mock.Mock()
        on_existing_event = mock.Mock(is_enabled=False, callback=mock.Mock(__name__="testing"))
        event_manager._consumers = {"existing_event": on_existing_event}
        shard = object()
        payload = {"berp": "baz"}

        event_manager.consume_raw_event("EXISTING_EVENT", shard, payload)

        event_manager.dispatch.assert_not_called()
        event_manager._event_factory.deserialize_shard_payload_event.assert_not_called()
        event_manager._enabled_for_event.assert_called_once_with(shard_events.ShardPayloadEvent)

    @pytest.mark.asyncio
    async def test_consume_raw_event_handles_exceptions(self, event_manager):
        mock_task = mock.Mock()
        # On Python 3.12+ Asyncio uses this to get the task's context if set to call the
        # error handler in. We want to avoid for this test for simplicity.
        mock_task.get_context.return_value = None
        exc = Exception("aaaa!")
        consumer = mock.Mock(callback=mock.Mock(__name__="testing", side_effect=exc))
        event_manager._consumers = {"existing_event": consumer}

        error_handler = mock.MagicMock()
        event_loop = asyncio.get_running_loop()
        event_loop.set_exception_handler(error_handler)
        shard = object()
        pl = {"i like": "cats"}

        with mock.patch.object(asyncio, "current_task", return_value=mock_task):
            event_manager.consume_raw_event("EXISTING_EVENT", shard, pl)

        error_handler.assert_called_once_with(
            event_loop,
            {
                "exception": exc,
                "message": "Exception occurred in raw event dispatch conduit",
                "payload": pl,
                "task": mock_task,
            },
        )

    def test_subscribe_when_class_call(self, event_manager):
        class Foo:
            async def __call__(self) -> None: ...

        foo = Foo()
        event_manager._check_event = mock.Mock()

        event_manager.subscribe(member_events.MemberCreateEvent, foo)

        assert event_manager._listeners[member_events.MemberCreateEvent] == [foo]

    def test_subscribe_when_callback_is_not_coroutine(self, event_manager):
        def test(): ...

        with pytest.raises(TypeError, match=r"Cannot subscribe a non-coroutine function callback"):
            event_manager.subscribe(member_events.MemberCreateEvent, test)

    def test_subscribe_when_event_type_not_in_listeners(self, event_manager):
        async def test(): ...

        event_manager._increment_listener_group_count = mock.Mock()
        event_manager._check_event = mock.Mock()

        event_manager.subscribe(member_events.MemberCreateEvent, test, _nested=1)

        assert event_manager._listeners == {member_events.MemberCreateEvent: [test]}
        event_manager._check_event.assert_called_once_with(member_events.MemberCreateEvent, 1)
        event_manager._increment_listener_group_count.assert_called_once_with(member_events.MemberCreateEvent, 1)

    def test_subscribe_when_event_type_in_listeners(self, event_manager):
        async def test(): ...

        async def test2(): ...

        event_manager._increment_listener_group_count = mock.Mock()
        event_manager._listeners[member_events.MemberCreateEvent] = [test2]
        event_manager._check_event = mock.Mock()

        event_manager.subscribe(member_events.MemberCreateEvent, test, _nested=2)

        assert event_manager._listeners == {member_events.MemberCreateEvent: [test2, test]}
        event_manager._check_event.assert_called_once_with(member_events.MemberCreateEvent, 2)
        event_manager._increment_listener_group_count.assert_not_called()

    @pytest.mark.parametrize("obj", ["test", event_manager_base.EventManagerBase])
    def test__check_event_when_event_type_does_not_subclass_Event(self, event_manager, obj):
        with pytest.raises(TypeError, match=r"'event_type' is a non-Event type"):
            event_manager._check_event(obj, 0)

    def test__check_event_when_no_intents_required(self, event_manager):
        event_manager._intents = intents.Intents.ALL

        with mock.patch.object(base_events, "get_required_intents_for", return_value=None) as get_intents:
            with mock.patch.object(warnings, "warn") as warn:
                event_manager._check_event(member_events.MemberCreateEvent, 0)

        get_intents.assert_called_once_with(member_events.MemberCreateEvent)
        warn.assert_not_called()

    def test__check_event_when_generic_event(self, event_manager):
        T = typing.TypeVar("T")

        class GenericEvent(typing.Generic[T], base_events.Event): ...

        event_manager._intents = intents.Intents.GUILD_MEMBERS

        with mock.patch.object(
            base_events, "get_required_intents_for", return_value=intents.Intents.GUILD_MEMBERS
        ) as get_intents:
            with mock.patch.object(warnings, "warn") as warn:
                event_manager._check_event(GenericEvent[int], 0)

        get_intents.assert_called_once_with(GenericEvent)
        warn.assert_not_called()

    def test__check_event_when_intents_correct(self, event_manager):
        event_manager._intents = intents.Intents.GUILD_EMOJIS | intents.Intents.GUILD_MEMBERS

        with mock.patch.object(
            base_events, "get_required_intents_for", return_value=intents.Intents.GUILD_MEMBERS
        ) as get_intents:
            with mock.patch.object(warnings, "warn") as warn:
                event_manager._check_event(member_events.MemberCreateEvent, 0)

        get_intents.assert_called_once_with(member_events.MemberCreateEvent)
        warn.assert_not_called()

    def test__check_event_when_intents_incorrect(self, event_manager):
        event_manager._intents = intents.Intents.GUILD_EMOJIS

        with mock.patch.object(
            base_events, "get_required_intents_for", return_value=intents.Intents.GUILD_MEMBERS
        ) as get_intents:
            with mock.patch.object(warnings, "warn") as warn:
                event_manager._check_event(member_events.MemberCreateEvent, 0)

        get_intents.assert_called_once_with(member_events.MemberCreateEvent)
        warn.assert_called_once_with(
            "You have tried to listen to MemberCreateEvent, but this will only ever be triggered if "
            "you enable one of the following intents: GUILD_MEMBERS.",
            category=errors.MissingIntentWarning,
            stacklevel=3,
        )

    def test_get_listeners_when_not_event(self, event_manager):
        event_manager._listeners = {}

        assert event_manager.get_listeners(base_events.Event) == []

    def test_get_listeners_polymorphic(self, event_manager):
        event_manager._listeners = {
            base_events.Event: ["coroutine0"],
            member_events.MemberEvent: ["coroutine1"],
            member_events.MemberCreateEvent: ["hi", "i am"],
            member_events.MemberUpdateEvent: ["hidden"],
            base_events.ExceptionEvent: ["so you won't see me"],
        }

        assert event_manager.get_listeners(member_events.MemberEvent) == ["coroutine1", "coroutine0"]

    def test_get_listeners_monomorphic_and_no_results(self, event_manager):
        event_manager._listeners = {
            member_events.MemberCreateEvent: ["coroutine1", "coroutine2"],
            member_events.MemberUpdateEvent: ["coroutine3"],
            member_events.MemberDeleteEvent: ["coroutine4", "coroutine5"],
        }

        assert event_manager.get_listeners(member_events.MemberEvent, polymorphic=False) == ()

    def test_get_listeners_monomorphic_and_results(self, event_manager):
        event_manager._listeners = {
            member_events.MemberEvent: ["coroutine0"],
            member_events.MemberCreateEvent: ["coroutine1", "coroutine2"],
            member_events.MemberUpdateEvent: ["coroutine3"],
            member_events.MemberDeleteEvent: ["coroutine4", "coroutine5"],
        }

        assert event_manager.get_listeners(member_events.MemberEvent, polymorphic=False) == ["coroutine0"]

    def test_unsubscribe_when_event_type_not_in_listeners(self, event_manager):
        async def test(): ...

        event_manager._increment_listener_group_count = mock.Mock()
        event_manager._listeners = {}

        event_manager.unsubscribe(member_events.MemberCreateEvent, test)

        assert event_manager._listeners == {}
        event_manager._increment_listener_group_count.assert_not_called()

    def test_unsubscribe_when_event_type_when_list_not_empty_after_delete(self, event_manager):
        async def test(): ...

        async def test2(): ...

        event_manager._increment_listener_group_count = mock.Mock()
        event_manager._listeners = {
            member_events.MemberCreateEvent: [test, test2],
            member_events.MemberDeleteEvent: [test],
        }

        event_manager.unsubscribe(member_events.MemberCreateEvent, test)

        assert event_manager._listeners == {
            member_events.MemberCreateEvent: [test2],
            member_events.MemberDeleteEvent: [test],
        }
        event_manager._increment_listener_group_count.assert_not_called()

    def test_unsubscribe_when_event_type_when_list_empty_after_delete(self, event_manager):
        async def test(): ...

        event_manager._increment_listener_group_count = mock.Mock()
        event_manager._listeners = {member_events.MemberCreateEvent: [test], member_events.MemberDeleteEvent: [test]}

        event_manager.unsubscribe(member_events.MemberCreateEvent, test)

        assert event_manager._listeners == {member_events.MemberDeleteEvent: [test]}
        event_manager._increment_listener_group_count.assert_called_once_with(member_events.MemberCreateEvent, -1)

    def test_listen_when_no_params(self, event_manager):
        with pytest.raises(TypeError):

            @event_manager.listen()
            async def test(): ...

    def test_listen_when_more_then_one_param_when_provided_in_typehint(self, event_manager):
        with pytest.raises(TypeError):

            @event_manager.listen()
            async def test(a, b, c): ...

    def test_listen_when_more_then_one_param_when_provided_in_decorator(self, event_manager):
        with pytest.raises(TypeError):

            @event_manager.listen(object)
            async def test(a, b, c): ...

    def test_listen_when_param_not_provided_in_decorator_nor_typehint(self, event_manager):
        with pytest.raises(TypeError):

            @event_manager.listen()
            async def test(event): ...

    def test_listen_when_param_provided_in_decorator(self, event_manager):
        stack = contextlib.ExitStack()

        subscribe = stack.enter_context(mock.patch.object(event_manager_base.EventManagerBase, "subscribe"))
        resolve_signature = stack.enter_context(mock.patch.object(reflect, "resolve_signature"))

        with stack:

            @event_manager.listen(member_events.MemberCreateEvent)
            async def test(event): ...

        resolve_signature.assert_not_called()
        subscribe.assert_called_once_with(member_events.MemberCreateEvent, test, _nested=1)

    def test_listen_when_multiple_params_provided_in_decorator(self, event_manager):
        stack = contextlib.ExitStack()

        subscribe = stack.enter_context(mock.patch.object(event_manager_base.EventManagerBase, "subscribe"))
        resolve_signature = stack.enter_context(mock.patch.object(reflect, "resolve_signature"))

        with stack:

            @event_manager.listen(member_events.MemberCreateEvent, member_events.MemberDeleteEvent)
            async def test(event): ...

        assert subscribe.call_count == 2
        resolve_signature.assert_not_called()
        subscribe.assert_has_calls(
            [
                mock.call(member_events.MemberCreateEvent, test, _nested=1),
                mock.call(member_events.MemberDeleteEvent, test, _nested=1),
            ]
        )

    def test_listen_when_param_provided_in_typehint(self, event_manager):
        with mock.patch.object(event_manager_base.EventManagerBase, "subscribe") as subscribe:

            @event_manager.listen()
            async def test(event: member_events.MemberCreateEvent): ...

        subscribe.assert_called_once_with(member_events.MemberCreateEvent, test, _nested=1)

    def test_listen_when_multiple_params_provided_as_typing_union_in_typehint(self, event_manager):
        with mock.patch.object(event_manager_base.EventManagerBase, "subscribe") as subscribe:

            @event_manager.listen()
            async def test(event: typing.Union[member_events.MemberCreateEvent, member_events.MemberDeleteEvent]): ...

        assert subscribe.call_count == 2
        subscribe.assert_has_calls(
            [
                mock.call(member_events.MemberCreateEvent, test, _nested=1),
                mock.call(member_events.MemberDeleteEvent, test, _nested=1),
            ]
        )

    @pytest.mark.skipif(sys.version_info < (3, 10), reason="Bitwise union only available on 3.10+")
    def test_listen_when_multiple_params_provided_as_bitwise_union_in_typehint(self, event_manager):
        with mock.patch.object(event_manager_base.EventManagerBase, "subscribe") as subscribe:

            @event_manager.listen()
            async def test(event: member_events.MemberCreateEvent | member_events.MemberDeleteEvent): ...

        assert subscribe.call_count == 2
        subscribe.assert_has_calls(
            [
                mock.call(member_events.MemberCreateEvent, test, _nested=1),
                mock.call(member_events.MemberDeleteEvent, test, _nested=1),
            ]
        )

    def test_listen_when_incorrect_type_in_typehint(self, event_manager):
        with pytest.raises(TypeError):

            @event_manager.listen()
            async def test(event: list[member_events.MemberUpdateEvent]): ...



# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Model that represents a common RGB color and provides simple conversions to other common color systems."""

from __future__ import annotations

__all__: typing.Sequence[str] = ("Color", "Colorish")

import re
import string
import typing

from hikari.internal import typing_extensions


def _to_rgb_int(value: str, name: str) -> int:
    # Heavy validation that is user-friendly and doesn't allow exploiting overflows, etc easily.
    #
    # isdigit allows chars like ² according to the docs.
    if not all(c in string.digits for c in value):
        msg = f"Expected digits only for {name} channel"
        raise ValueError(msg)
    if not value or len(value) > 3:
        msg = f"Expected 1 to 3 digits for {name} channel, got {len(value)}"
        raise ValueError(msg)

    int_value = int(value)

    if int_value >= 256:
        msg = f"Expected {name} channel to be in the inclusive range of 0 and 255, got {value!r}"
        raise ValueError(msg)

    return int_value


_FLOAT_PATTERN: typing.Final[typing.Pattern[str]] = re.compile(r"0\.\d*|\.\d+|1\.0*")


def _to_rgb_float(value: str, name: str) -> float:
    # Floats are easier to handle, as they don't overflow, they just become [`inf`][].

    if value.count(".") != 1:
        msg = f'Expected exactly 1 decimal point "." in {name} channel'
        raise ValueError(msg)
    if not _FLOAT_PATTERN.match(value):
        msg = f"Expected {name} channel to be a decimal in the inclusive range of 0.0 and 1.0"
        raise ValueError(msg)
    return float(value)


class Color(int):
    """Representation of a color.

    This value is immutable.

    This is a specialization of [`int`][] which provides alternative overrides for
    common methods and color system conversions.

    This currently supports:

    * RGB
    * RGB (float)
    * 3-digit hex codes (e.g. 0xF1A -- web safe)
    * 6-digit hex codes (e.g. 0xFF11AA)
    * 3-digit RGB strings (e.g. #1A2 -- web safe)
    * 6-digit RGB hash strings (e.g. #1A2B3C)

    Examples
    --------
    Examples of conversions to given formats include:

    ```py
    >>> c = Color(0xFF051A)
    Color(r=0xff, g=0x5, b=0x1a)

    >>> hex(c)
    0xff051a

    >>> c.hex_code
    #FF051A

    >>> str(c)
    #FF051A

    >>> int(c)
    16712986

    >>> c.rgb
    (255, 5, 26)

    >>> c.rgb_float
    (1.0, 0.0196078431372549, 0.10196078431372549)
    ```

    Alternatively, if you have an arbitrary input in one of the above formats
    that you wish to become a color, you can use [`hikari.colors.Color.of`][] on
    the class itself to automatically attempt to resolve the color:

    ```py
    >>> Color.of(0xFF051A)
    Color(r=0xff, g=0x5, b=0x1a)

    >>> Color.of(16712986)
    Color(r=0xff, g=0x5, b=0x1a)

    >>> c = Color.of((255, 5, 26))
    Color(r=0xff, g=0x5, b=1xa)

    >>> c = Color.of(255, 5, 26)
    Color(r=0xff, g=0x5, b=1xa)

    >>> c = Color.of([0xFF, 0x5, 0x1a])
    Color(r=0xff, g=0x5, b=1xa)

    >>> c = Color.of("#1a2b3c")
    Color(r=0x1a, g=0x2b, b=0x3c)

    >>> c = Color.of("#1AB")
    Color(r=0x11, g=0xaa, b=0xbb)

    >>> c = Color.of((1.0, 0.0196078431372549, 0.10196078431372549))
    Color(r=0xff, g=0x5, b=0x1a)

    >>> c = Color.of([1.0, 0.0196078431372549, 0.10196078431372549])
    Color(r=0xff, g=0x5, b=0x1a)
    ```

    Examples of initialization of Color objects from given formats include:

    ```py
    >>> c = Color(16712986)
    Color(r=0xff, g=0x5, b=0x1a)

    >>> c = Color.from_rgb(255, 5, 26)
    Color(r=0xff, g=0x5, b=1xa)

    >>> c = Color.from_hex_code("#1a2b3c")
    Color(r=0x1a, g=0x2b, b=0x3c)

    >>> c = Color.from_hex_code("#1AB")
    Color(r=0x11, g=0xaa, b=0xbb)

    >>> c = Color.from_rgb_float(1.0, 0.0196078431372549, 0.10196078431372549)
    Color(r=0xff, g=0x5, b=0x1a)
    ```
    """

    __slots__: typing.Sequence[str] = ()

    def __init__(self, raw_rgb: typing.SupportsInt) -> None:
        if not (0 <= int(raw_rgb) <= 0xFFFFFF):
            msg = f"raw_rgb must be in the exclusive range of 0 and {0xFF_FF_FF}"
            raise ValueError(msg)
        # The __new__ for [`int`][] initializes the value for us, this super-call does nothing other
        # than keeping the linter happy.
        super().__init__()

    @typing_extensions.override
    def __repr__(self) -> str:
        r, g, b = self.rgb
        return f"Color(r={hex(r)}, g={hex(g)}, b={hex(b)})"

    @typing_extensions.override
    def __str__(self) -> str:
        return self.hex_code

    @property
    def rgb(self) -> tuple[int, int, int]:
        """The RGB representation of this Color.

        Represented as a tuple of R, G, B. Each value is
        in the range [0, 0xFF].

        Examples
        --------
        `(123, 234, 47)`
        """
        return (self >> 16) & 0xFF, (self >> 8) & 0xFF, self & 0xFF

    @property
    def rgb_float(self) -> tuple[float, float, float]:
        """Return the floating-point RGB representation of this Color.

        Represented as a tuple of R, G, B. Each value is in the range [0, 1].

        Examples
        --------
        `(0.1, 0.2, 0.76)`
        """
        r, g, b = self.rgb
        return r / 0xFF, g / 0xFF, b / 0xFF

    @property
    def hex_code(self) -> str:
        """Six-digit hexadecimal color code for this Color.

        This is prepended with a `#` symbol, and will be in upper case.

        Examples
        --------
        `#1A2B3C`
        """
        return "#" + self.raw_hex_code

    @property
    def raw_hex_code(self) -> str:
        """Raw hex code.

        Examples
        --------
        `1A2B3C`
        """
        components = self.rgb
        return "".join(hex(c)[2:].zfill(2) for c in components).upper()

    @property
    def is_web_safe(self) -> bool:
        """Whether this color is web safe."""
        return not (((self & 0xFF0000) % 0x110000) or ((self & 0xFF00) % 0x1100) or ((self & 0xFF) % 0x11))

    @classmethod
    def from_rgb(cls, red: int, green: int, blue: int, /) -> Color:
        """Convert the given RGB to a [`hikari.colors.Color`][] object.

        Each channel must be within the range [0, 255] (0x0, 0xFF).

        Parameters
        ----------
        red
            Red channel.
        green
            Green channel.
        blue
            Blue channel.

        Returns
        -------
        Color
            A Color object.

        Raises
        ------
        ValueError
            If red, green, or blue are outside the range [0x0, 0xFF].
        """
        if not 0 <= red <= 0xFF:
            msg = "Expected red channel to be in the inclusive range of 0 and 255"
            raise ValueError(msg)
        if not 0 <= green <= 0xFF:
            msg = "Expected green channel to be in the inclusive range of 0 and 255"
            raise ValueError(msg)
        if not 0 <= blue <= 0xFF:
            msg = "Expected blue channel to be in the inclusive range of 0 and 255"
            raise ValueError(msg)
        return cls((red << 16) | (green << 8) | blue)

    @classmethod
    def from_rgb_float(cls, red: float, green: float, blue: float, /) -> Color:
        """Convert the given RGB to a [`hikari.colors.Color`][] object.

        The color-space represented values have to be within the
        range [0, 1].

        Parameters
        ----------
        red
            Red channel.
        green
            Green channel.
        blue
            Blue channel.

        Returns
        -------
        Color
            A Color object.

        Raises
        ------
        ValueError
            If red, green or blue are outside the range [0, 1].
        """
        if not 0 <= red <= 1:
            msg = "Expected red channel to be in the inclusive range of 0.0 and 1.0"
            raise ValueError(msg)
        if not 0 <= green <= 1:
            msg = "Expected green channel to be in the inclusive range of 0.0 and 1.0"
            raise ValueError(msg)
        if not 0 <= blue <= 1:
            msg = "Expected blue channel to be in the inclusive range of 0.0 and 1.0"
            raise ValueError(msg)
        return cls.from_rgb(int(red * 0xFF), int(green * 0xFF), int(blue * 0xFF))

    @classmethod
    def from_hex_code(cls, hex_code: str, /) -> Color:
        """Convert the given hexadecimal color code to a [`hikari.colors.Color`][].

        The inputs may be of the following format (case insensitive):
        `1a2`, `#1a2`, `0x1a2` (for web-safe colors), or
        `1a2b3c`, `#1a2b3c`, `0x1a2b3c` (for regular 3-byte color-codes).

        Parameters
        ----------
        hex_code
            A hexadecimal color code to parse. This may optionally start with
            a case insensitive `0x` or `#`.

        Returns
        -------
        Color
            A corresponding Color object.

        Raises
        ------
        ValueError
            If `hex_code` is not a hexadecimal or is a invalid length.
        """
        if hex_code.startswith("#"):
            hex_code = hex_code[1:]
        elif hex_code.startswith(("0x", "0X")):
            hex_code = hex_code[2:]

        if not all(c in string.hexdigits for c in hex_code):
            msg = "Color code must be hexadecimal"
            raise ValueError(msg)

        if len(hex_code) == 3:
            # Web-safe
            r, g, b = (c << 4 | c for c in (int(c, 16) for c in hex_code))
            return cls.from_rgb(r, g, b)

        if len(hex_code) == 6:
            return cls.from_rgb(int(hex_code[:2], 16), int(hex_code[2:4], 16), int(hex_code[4:6], 16))

        msg = "Color code is invalid length. Must be 3 or 6 digits"
        raise ValueError(msg)

    @classmethod
    def from_int(cls, integer: typing.SupportsInt, /) -> Color:
        """Convert the given [`typing.SupportsInt`][] to a [`hikari.colors.Color`][].

        Parameters
        ----------
        integer
            The raw color integer.

        Returns
        -------
        Color
            The Color object.
        """
        return cls(integer)

    @classmethod
    def from_tuple_string(cls, tuple_str: str, /) -> Color:
        """Convert a string in a tuple-like format to a [`hikari.colors.Color`][].

        This allows formats that are optionally enclosed by `()`,
        `{}`, `[]` or `<>`, and contain three floats or ints,
        either space separated or comma separated.

        If comma separated, trailing and leading whitespace around each member
        is truncated.

        This is provided to allow command frontends to directly pass user
        input for representing a given colour into this class safely.

        Examples
        --------
        ```py
            # Floats
            "1.0 1.0 1.0"
            "(1.0 1.0 1.0)"
            "[1.0 1.0 1.0]"
            "{1.0 1.0 1.0}"
            "1.0, 1.0, 1.0"
            "(1.0, 1.0, 1.0)"
            "[1.0, 1.0, 1.0]"
            "{1.0, 1.0, 1.0}"

            # Ints
            "252 252 252"
            "(252 252 252)"
            "[252 252 252]"
            "{252 252 252}"
            "252, 252, 252"
            "(252, 252, 252)"
            "[252, 252, 252]"
            "{252, 252, 252}"
        ```

        Parameters
        ----------
        tuple_str
            The string to parse.

        Returns
        -------
        Color
            The parsed colour object.

        Raises
        ------
        ValueError
            If an invalid format is given, or if any values exceed 1.0 for
            floats or 255 for ints.
        """
        if tuple_str[:: len(tuple_str) - 1] in {"()", "{}", "<>", "[]"}:
            tuple_str = tuple_str[1:-1].strip()

        try:
            if "," in tuple_str:
                r, g, b = (bit.strip() for bit in tuple_str.split(","))
            else:
                r, g, b = tuple_str.split()
        except ValueError:
            msg = "Expected three comma/space separated values"
            raise ValueError(msg) from None

        if any("." in s for s in (r, g, b)):
            return cls.from_rgb_float(_to_rgb_float(r, "red"), _to_rgb_float(g, "green"), _to_rgb_float(b, "blue"))
        return cls.from_rgb(_to_rgb_int(r, "red"), _to_rgb_int(g, "green"), _to_rgb_int(b, "blue"))

    @classmethod
    def of(cls, value: Colorish, /) -> Color:
        """Convert the value to a [`hikari.colors.Color`][].

        This attempts to determine the correct data format based on the
        information provided.

        Parameters
        ----------
        value
            A color compatible values.

        Examples
        --------
        ```py
        >>> Color.of(0xFF051A)
        Color(r=0xff, g=0x5, b=0x1a)

        >>> Color.of(16712986)
        Color(r=0xff, g=0x5, b=0x1a)

        >>> c = Color.of((255, 5, 26))
        Color(r=0xff, g=0x5, b=1xa)

        >>> c = Color.of([0xFF, 0x5, 0x1a])
        Color(r=0xff, g=0x5, b=1xa)

        >>> c = Color.of("#1a2b3c")
        Color(r=0x1a, g=0x2b, b=0x3c)

        >>> c = Color.of("#1AB")
        Color(r=0x11, g=0xaa, b=0xbb)

        >>> c = Color.of((1.0, 0.0196078431372549, 0.10196078431372549))
        Color(r=0xff, g=0x5, b=0x1a)

        >>> c = Color.of([1.0, 0.0196078431372549, 0.10196078431372549])
        Color(r=0xff, g=0x5, b=0x1a)

        # Commas and brackets are optional, whitespace is ignored, and these
        # are compatible with all-ints between 0-255 or all-floats between
        # 0.0 and 1.0 only.
        >>> c = Color.of("5, 22, 33")
        Color(r=0x5, g=0x16, b=0x21)
        >>> c = Color.of("(5, 22, 33)")
        Color(r=0x5, g=0x16, b=0x21)
        >>> c = Color.of("[5, 22, 33]")
        Color(r=0x5, g=0x16, b=0x21)
        >>> c = Color.of("{5, 22, 33}")
        Color(r=0x5, g=0x16, b=0x21)
        ```

        Returns
        -------
        Color
            The Color object.
        """
        if isinstance(value, cls):
            return value
        if isinstance(value, int):
            return cls.from_int(value)
        if isinstance(value, (list, tuple)):
            if len(value) != 3:
                msg = f"Color must be an RGB triplet if set to a {type(value).__name__} type"
                raise ValueError(msg)

            r, g, b = value

            if isinstance(r, float) and isinstance(g, float) and isinstance(b, float):
                return cls.from_rgb_float(r, g, b)

            if isinstance(r, int) and isinstance(g, int) and isinstance(b, int):
                return cls.from_rgb(r, g, b)

        if isinstance(value, str):
            if any(c in value for c in "({[<,. "):
                return cls.from_tuple_string(value)

            is_start_hash_or_hex_literal = value.casefold().startswith(("#", "0x"))
            is_hex_digits = all(c in string.hexdigits for c in value) and len(value) in (3, 6)
            if is_start_hash_or_hex_literal or is_hex_digits:
                return cls.from_hex_code(value)

        msg = f"Could not transform {value!r} into a {cls.__qualname__} object"
        raise ValueError(msg)

    @typing_extensions.override
    def to_bytes(
        self, length: typing.SupportsIndex, byteorder: typing.Literal["little", "big"], *, signed: bool = True
    ) -> bytes:
        """Convert the color code to bytes.

        Parameters
        ----------
        length
            The number of bytes to produce. Should be around `3`, but not less.
        byteorder
            The endianness of the value represented by the bytes.
            Can be `"big"` endian or `"little"` endian.
        signed
            Whether the value is signed or unsigned.

        Returns
        -------
        bytes
            The bytes representation of the Color.
        """
        return int(self).to_bytes(length, byteorder, signed=signed)


Colorish = typing.Union[
    Color,
    typing.SupportsInt,
    tuple[typing.SupportsInt, typing.SupportsInt, typing.SupportsInt],
    tuple[typing.SupportsFloat, typing.SupportsFloat, typing.SupportsFloat],
    typing.Sequence[typing.SupportsInt],
    typing.Sequence[typing.SupportsFloat],
    str,
]
"""Type hint representing types of value compatible with a colour type.

This may be:

1. [`hikari.colors.Color`][]
2. [`hikari.colours.Colour`][] (an alias for [`hikari.colors.Color`][]).
3. A value that can be cast to an [`int`][] (RGB hex-code).
4. a 3-[`tuple`][] of [`int`][] (RGB integers in range 0 through 255).
5. a 3-[`tuple`][] of [`float`][] (RGB floats in range 0 through 1).
6. a list of [`int`][].
7. a list of [`float`][].
8. a [`str`][] hex colour code.

A hex colour code is expected to be in one of the following formats. Each of the
following examples means the same thing semantically.

1. (web-safe) `"12F"` (equivalent to `"1122FF"`)
2. (web-safe) `"0x12F"` (equivalent to `"0x1122FF"`)
3. (web-safe) `"0X12F"` (equivalent to `"0X1122FF"`)
4. (web-safe) `"#12F"` (equivalent to `"#1122FF"`)
5. `"1122FF"`
6. `"0x1122FF"`
7. `"0X1122FF"`
8. `"#1122FF"`

Web-safe colours are three hex-digits wide, `XYZ` becomes `XXYYZZ` in full-form.
"""



# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import datetime

import mock
import pytest

from hikari import emojis
from hikari import guilds
from hikari import messages
from hikari import snowflakes
from hikari import undefined
from hikari import urls
from hikari import users
from hikari.internal import routes


class TestAttachment:
    def test_str_operator(self):
        attachment = messages.Attachment(
            id=123,
            filename="super_cool_file.cool",
            title="other title",
            description="description!",
            media_type="image/png",
            height=222,
            width=555,
            proxy_url="htt",
            size=543,
            url="htttt",
            is_ephemeral=False,
            duration=1,
            waveform="122111",
        )
        assert str(attachment) == "super_cool_file.cool"


class TestReaction:
    def test_str_operator(self):
        reaction = messages.Reaction(emoji=emojis.UnicodeEmoji("\N{OK HAND SIGN}"), count=42, is_me=True)
        assert str(reaction) == "\N{OK HAND SIGN}"


class TestMessageApplication:
    @pytest.fixture
    def message_application(self):
        return messages.MessageApplication(
            id=123, name="test app", description="", icon_hash="123abc", cover_image_hash="abc123"
        )

    def test_make_cover_url_format_set_to_deprecated_ext_argument_if_provided(self, message_application):
        with mock.patch.object(
            routes, "CDN_APPLICATION_COVER", new=mock.Mock(compile_to_file=mock.Mock(return_value="file"))
        ) as route:
            assert message_application.make_cover_image_url(ext="JPEG") == "file"

        route.compile_to_file.assert_called_once_with(
            urls.CDN_URL, application_id=123, hash="abc123", size=4096, file_format="JPEG", lossless=True
        )

    def test_cover_image_url(self, message_application):
        with mock.patch.object(messages.MessageApplication, "make_cover_image_url") as mock_cover_image:
            assert message_application.cover_image_url is mock_cover_image()

    def test_make_cover_image_url_when_hash_is_none(self, message_application):
        message_application.cover_image_hash = None

        assert message_application.make_cover_image_url() is None

    def test_make_cover_image_url_when_hash_is_not_none(self, message_application):
        with mock.patch.object(
            routes, "CDN_APPLICATION_COVER", new=mock.Mock(compile_to_file=mock.Mock(return_value="file"))
        ) as route:
            assert message_application.make_cover_image_url(file_format="JPEG", size=1000) == "file"

        route.compile_to_file.assert_called_once_with(
            urls.CDN_URL, application_id=123, hash="abc123", size=1000, file_format="JPEG", lossless=True
        )


@pytest.fixture
def message():
    return messages.Message(
        app=None,
        id=snowflakes.Snowflake(1234),
        channel_id=snowflakes.Snowflake(5678),
        guild_id=snowflakes.Snowflake(910112),
        author=mock.Mock(spec_set=users.User),
        member=mock.Mock(spec_set=guilds.Member),
        content="blahblahblah",
        timestamp=datetime.datetime.now().astimezone(),
        edited_timestamp=None,
        is_tts=False,
        user_mentions={},
        role_mention_ids=[],
        channel_mentions={},
        mentions_everyone=False,
        attachments=(),
        embeds=(),
        poll=object(),
        reactions=(),
        is_pinned=True,
        webhook_id=None,
        type=messages.MessageType.DEFAULT,
        activity=None,
        application=None,
        message_reference=None,
        message_snapshots=None,
        flags=None,
        nonce=None,
        referenced_message=None,
        stickers=[],
        application_id=123123,
        components=[],
        thread=None,
        interaction_metadata=None,
    )


class TestMessage:
    def test_make_link_when_guild_is_not_none(self, message):
        message.id = 789
        message.channel_id = 456
        assert message.make_link(123) == "https://discord.com/channels/123/456/789"

    def test_make_link_when_guild_is_none(self, message):
        message.app = mock.Mock()
        message.id = 789
        message.channel_id = 456
        assert message.make_link(None) == "https://discord.com/channels/@me/456/789"


@pytest.fixture
def message_reference():
    return messages.MessageReference(
        app=None,
        guild_id=snowflakes.Snowflake(123),
        channel_id=snowflakes.Snowflake(456),
        id=snowflakes.Snowflake(789),
        type=messages.MessageReferenceType.DEFAULT,
    )


class TestMessageReference:
    def test_make_link_when_guild_is_not_none(self, message_reference):
        assert message_reference.message_link == "https://discord.com/channels/123/456/789"
        assert message_reference.channel_link == "https://discord.com/channels/123/456"

    def test_make_link_when_guild_is_none(self, message_reference):
        message_reference.guild_id = None
        assert message_reference.message_link == "https://discord.com/channels/@me/456/789"
        assert message_reference.channel_link == "https://discord.com/channels/@me/456"

    def test_make_link_when_id_is_none(self, message_reference):
        message_reference.id = None
        assert message_reference.message_link is None
        assert message_reference.channel_link == "https://discord.com/channels/123/456"


@pytest.mark.asyncio
class TestAsyncMessage:
    async def test_fetch_channel(self, message):
        message.app = mock.AsyncMock()
        message.channel_id = 123
        await message.fetch_channel()
        message.app.rest.fetch_channel.assert_awaited_once_with(123)

    async def test_edit(self, message):
        message.app = mock.AsyncMock()
        message.id = 123
        message.channel_id = 456
        embed = object()
        embeds = [object(), object()]
        component = object()
        components = object(), object()
        attachment = object()
        roles = [object()]
        await message.edit(
            content="test content",
            embed=embed,
            embeds=embeds,
            attachment=attachment,
            attachments=[attachment, attachment],
            component=component,
            components=components,
            mentions_everyone=True,
            mentions_reply=False,
            user_mentions=False,
            role_mentions=roles,
            flags=messages.MessageFlag.URGENT,
        )
        message.app.rest.edit_message.assert_awaited_once_with(
            message=123,
            channel=456,
            content="test content",
            embed=embed,
            embeds=embeds,
            attachment=attachment,
            attachments=[attachment, attachment],
            component=component,
            components=components,
            mentions_everyone=True,
            mentions_reply=False,
            user_mentions=False,
            role_mentions=roles,
            flags=messages.MessageFlag.URGENT,
        )

    async def test_respond(self, message):
        message.app = mock.AsyncMock()
        message.id = 123
        message.channel_id = 456
        embed = object()
        embeds = [object(), object()]
        poll = object()
        roles = [object()]
        attachment = object()
        attachments = [object()]
        component = object()
        components = object(), object()
        reference_messsage = object()
        await message.respond(
            content="test content",
            embed=embed,
            embeds=embeds,
            poll=poll,
            attachment=attachment,
            attachments=attachments,
            component=component,
            components=components,
            sticker=123,
            stickers=[543, 6542],
            tts=True,
            reply=reference_messsage,
            reply_must_exist=False,
            mentions_everyone=True,
            user_mentions=False,
            role_mentions=roles,
            mentions_reply=True,
            flags=321123,
        )
        message.app.rest.create_message.assert_awaited_once_with(
            channel=456,
            content="test content",
            embed=embed,
            embeds=embeds,
            poll=poll,
            attachment=attachment,
            attachments=attachments,
            component=component,
            components=components,
            sticker=123,
            stickers=[543, 6542],
            tts=True,
            reply=reference_messsage,
            reply_must_exist=False,
            mentions_everyone=True,
            user_mentions=False,
            role_mentions=roles,
            mentions_reply=True,
            flags=321123,
        )

    async def test_respond_when_reply_is_True(self, message):
        message.app = mock.AsyncMock()
        message.id = 123
        message.channel_id = 456
        await message.respond(reply=True)
        message.app.rest.create_message.assert_awaited_once_with(
            channel=456,
            content=undefined.UNDEFINED,
            embed=undefined.UNDEFINED,
            embeds=undefined.UNDEFINED,
            poll=undefined.UNDEFINED,
            attachment=undefined.UNDEFINED,
            attachments=undefined.UNDEFINED,
            component=undefined.UNDEFINED,
            components=undefined.UNDEFINED,
            sticker=undefined.UNDEFINED,
            stickers=undefined.UNDEFINED,
            tts=undefined.UNDEFINED,
            reply=message,
            reply_must_exist=undefined.UNDEFINED,
            mentions_everyone=undefined.UNDEFINED,
            user_mentions=undefined.UNDEFINED,
            role_mentions=undefined.UNDEFINED,
            mentions_reply=undefined.UNDEFINED,
            flags=undefined.UNDEFINED,
        )

    async def test_respond_when_reply_is_False(self, message):
        message.app = mock.AsyncMock()
        message.id = 123
        message.channel_id = 456
        await message.respond(reply=False)
        message.app.rest.create_message.assert_awaited_once_with(
            channel=456,
            content=undefined.UNDEFINED,
            embed=undefined.UNDEFINED,
            embeds=undefined.UNDEFINED,
            poll=undefined.UNDEFINED,
            attachment=undefined.UNDEFINED,
            attachments=undefined.UNDEFINED,
            component=undefined.UNDEFINED,
            components=undefined.UNDEFINED,
            sticker=undefined.UNDEFINED,
            stickers=undefined.UNDEFINED,
            tts=undefined.UNDEFINED,
            reply=undefined.UNDEFINED,
            reply_must_exist=undefined.UNDEFINED,
            mentions_everyone=undefined.UNDEFINED,
            user_mentions=undefined.UNDEFINED,
            role_mentions=undefined.UNDEFINED,
            mentions_reply=undefined.UNDEFINED,
            flags=undefined.UNDEFINED,
        )

    async def test_delete(self, message):
        message.app = mock.AsyncMock()
        message.id = 123
        message.channel_id = 456
        await message.delete()
        message.app.rest.delete_message.assert_awaited_once_with(456, 123)

    async def test_add_reaction(self, message):
        message.app = mock.AsyncMock()
        message.id = 123
        message.channel_id = 456
        await message.add_reaction("👌", 123123)
        message.app.rest.add_reaction.assert_awaited_once_with(channel=456, message=123, emoji="👌", emoji_id=123123)

    async def test_remove_reaction(self, message):
        message.app = mock.AsyncMock()
        message.id = 123
        message.channel_id = 456
        await message.remove_reaction("👌", 341231)
        message.app.rest.delete_my_reaction.assert_awaited_once_with(
            channel=456, message=123, emoji="👌", emoji_id=341231
        )

    async def test_remove_reaction_with_user(self, message):
        message.app = mock.AsyncMock()
        user = object()
        message.id = 123
        message.channel_id = 456
        await message.remove_reaction("👌", 31231, user=user)
        message.app.rest.delete_reaction.assert_awaited_once_with(
            channel=456, message=123, emoji="👌", emoji_id=31231, user=user
        )

    async def test_remove_all_reactions(self, message):
        message.app = mock.AsyncMock()
        message.id = 123
        message.channel_id = 456
        await message.remove_all_reactions()
        message.app.rest.delete_all_reactions.assert_awaited_once_with(channel=456, message=123)

    async def test_remove_all_reactions_with_emoji(self, message):
        message.app = mock.AsyncMock()
        message.id = 123
        message.channel_id = 456
        await message.remove_all_reactions("👌", emoji_id=65655)
        message.app.rest.delete_all_reactions_for_emoji.assert_awaited_once_with(
            channel=456, message=123, emoji="👌", emoji_id=65655
        )



# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import array as array_
import operator
import sys

import mock
import pytest

from hikari.internal import collections


class TestFreezableDict:
    def test___init___with_source(self):
        mock_map = collections.FreezableDict({"o": "NO", "bye": "blam", "foo": "bar"})
        assert mock_map == {"o": "NO", "bye": "blam", "foo": "bar"}

    def test_clear(self):
        mock_map = collections.FreezableDict({"foo": "bar", "crash": "balloon"})
        mock_map.clear()

        assert mock_map._data == {}

    def test_copy(self):
        mock_map = collections.FreezableDict({"foo": "bar", "crash": "balloon"})
        result = mock_map.copy()

        assert result == {"foo": "bar", "crash": "balloon"}
        assert isinstance(result, collections.FreezableDict)
        assert result is not mock_map

    def test_freeze(self):
        mock_map = collections.FreezableDict({"hikari": "shinji", "gendo": "san"})
        result = mock_map.freeze()

        assert result == {"hikari": "shinji", "gendo": "san"}
        assert isinstance(result, dict)

    def test___delitem__(self):
        mock_map = collections.FreezableDict({"hikari": "shinji", "gendo": "san", "screwed": "up"})
        del mock_map["hikari"]

        assert mock_map == {"gendo": "san", "screwed": "up"}

    def test___getitem__(self):
        mock_map = collections.FreezableDict({"curiosity": "rover", "ok": "bye"})
        assert mock_map["ok"] == "bye"

    def test____iter__(self):
        mock_map = collections.FreezableDict({"curiosity": "rover", "cat": "bag", "ok": "bye"})
        assert list(mock_map) == ["curiosity", "cat", "ok"]

    def test___len__(self):
        mock_map = collections.FreezableDict({"hmm": "blam", "cat": "bag", "ok": "bye"})
        assert len(mock_map) == 3

    def test___setitem__(self):
        mock_map = collections.FreezableDict({"hmm": "forearm", "cat": "bag", "ok": "bye"})
        mock_map["bye"] = 4

        assert mock_map == {"hmm": "forearm", "cat": "bag", "ok": "bye", "bye": 4}


class TestLimitedCapacityCacheMap:
    def test___init___with_source(self):
        raw_map = {"voo": "doo", "blam": "blast", "foo": "bye"}
        mock_map = collections.LimitedCapacityCacheMap(raw_map, limit=2)
        assert mock_map == {"blam": "blast", "foo": "bye"}

    def test_clear(self):
        mock_map = collections.LimitedCapacityCacheMap({"o": "n", "b": "a", "a": "v"}, limit=42)
        mock_map.clear()

        assert mock_map._data == {}

    def test_copy(self):
        mock_map = collections.LimitedCapacityCacheMap({"o": "n", "b": "a", "a": "v"}, limit=42)
        result = mock_map.copy()

        assert result is not mock_map
        assert isinstance(result, collections.LimitedCapacityCacheMap)
        assert result == {"o": "n", "b": "a", "a": "v"}

    def test_freeze(self):
        mock_map = collections.LimitedCapacityCacheMap({"o": "no", "good": "bye"}, limit=5)
        result = mock_map.freeze()

        assert isinstance(result, dict)
        assert result == {"o": "no", "good": "bye"}

    def test___delitem___for_existing_entry(self):
        mock_map = collections.LimitedCapacityCacheMap(limit=50)
        mock_map["Ok"] = 42
        del mock_map["Ok"]
        assert "Ok" not in mock_map

    def test___delitem___for_non_existing_entry(self):
        mock_map = collections.LimitedCapacityCacheMap(limit=50)
        with pytest.raises(KeyError):
            del mock_map["Blam"]

    def test___getitem___for_existing_entry(self):
        mock_map = collections.LimitedCapacityCacheMap(limit=50)
        mock_map["blat"] = 42
        assert mock_map["blat"] == 42

    def test___getitem___for_non_existing_entry(self):
        mock_map = collections.LimitedCapacityCacheMap(limit=50)
        with pytest.raises(KeyError):
            mock_map["CIA"]

    def test___iter___(self):
        mock_map = collections.LimitedCapacityCacheMap(limit=50)
        mock_map.update({"OK": "blam", "blaaa": "neoeo", "neon": "genesis", "evangelion": None})
        assert list(mock_map) == ["OK", "blaaa", "neon", "evangelion"]

    def test___len___(self):
        mock_map = collections.LimitedCapacityCacheMap(limit=50)
        mock_map.update({"ooga": "blam", "blaaa": "neoeo", "the": "boys", "neon": "genesis", "evangelion": None})
        assert len(mock_map) == 5

    def test___setitem___when_limit_not_reached(self):
        mock_map = collections.LimitedCapacityCacheMap(limit=50)
        mock_map["OK"] = 523
        mock_map["blam"] = 512387
        mock_map.update({"bll": "no", "ieiei": "lslsl"})
        assert mock_map == {"OK": 523, "blam": 512387, "bll": "no", "ieiei": "lslsl"}

    def test___setitem___when_limit_reached(self):
        mock_map = collections.LimitedCapacityCacheMap(limit=4)
        mock_map.update({"bll": "no", "ieiei": "lslsl", "pacify": "me", "qt": "pie"})
        mock_map["eva"] = "Rei"
        mock_map.update({"shinji": "ikari"})
        assert mock_map == {"pacify": "me", "qt": "pie", "eva": "Rei", "shinji": "ikari"}

    def test___setitem___when_limit_reached_and_expire_callback_set(self):
        expire_callback = mock.Mock()
        mock_map = collections.LimitedCapacityCacheMap(limit=4, on_expire=expire_callback)
        mock_map.update({"bll": "no", "ieiei": "lslsl", "pacify": "me", "qt": "pie"})
        mock_map["eva"] = "Rei"
        mock_map.update({"shinji": "ikari"})
        expire_callback.assert_has_calls((mock.call("no"), mock.call("lslsl")))


class TestSnowflakeSet:
    def test_init_creates_empty_array(self):
        # given
        array = collections.SnowflakeSet()._ids
        # then
        assert len(array) == 0, "not empty"
        assert isinstance(array, array_.array)

    def test_init_creates_array(self):
        # given
        array = collections.SnowflakeSet(9, 8, 7, 8)._ids
        # then
        assert len(array) == 3, "wrong size"
        assert isinstance(array, array_.array)
        assert array.tolist() == [7, 8, 9]

    @pytest.mark.parametrize(
        ("start_with", "add_items", "expect"),
        [
            ([], [123], [123]),
            ([123], [123], [123]),
            ([123], [122], [122, 123]),
            ([123], [124], [123, 124]),
            ([0, 99999], [124], [0, 124, 99999]),
            ([0, 124, 99999], [124], [0, 124, 99999]),
            ([0, 122, 124, 99999], [122], [0, 122, 124, 99999]),
            ([0, 122, 124, 99999], [0, 123, 121], [0, 121, 122, 123, 124, 99999]),
            ([0, 122], [123, 121, 999991, 121, 121, 124, 120], [0, 120, 121, 122, 123, 124, 999991]),
        ],
    )
    def test_add_inserts_items(self, start_with, add_items, expect):
        # given
        sfs = collections.SnowflakeSet()
        sfs._ids.extend(start_with)
        # when
        for item in add_items:
            sfs.add(item)
        # then
        assert sfs._ids.tolist() == expect

    @pytest.mark.parametrize(
        ("start_with", "add_items", "expect"),
        [
            ([], [123], [123]),
            ([123], [123], [123]),
            ([123], [122], [122, 123]),
            ([123], [124], [123, 124]),
            ([0, 99999], [124], [0, 124, 99999]),
            ([0, 124, 99999], [124], [0, 124, 99999]),
            ([0, 122, 124, 99999], [122], [0, 122, 124, 99999]),
            ([0, 122, 124, 99999], [0, 123, 121], [0, 121, 122, 123, 124, 99999]),
            ([0, 122], [123, 121, 999991, 121, 121, 124, 120], [0, 120, 121, 122, 123, 124, 999991]),
        ],
    )
    def test_add_all_inserts_items(self, start_with, add_items, expect):
        # given
        sfs = collections.SnowflakeSet()
        sfs._ids.extend(start_with)
        # when
        sfs.add_all(add_items)
        # then
        assert sfs._ids.tolist() == expect

    def test_add_all_when_empty_collection_passed(self):
        started = [0, 122]
        # given
        sfs = collections.SnowflakeSet()
        sfs._ids.extend(started)
        # when
        sfs.add_all([])
        # then
        assert sfs._ids.tolist() == started

    def test_clear_empties_buffer(self):
        # given
        sfs = collections.SnowflakeSet(123, 121, 999991, 121, 121, 124, 120)
        assert len(sfs._ids) != 0
        # when
        sfs.clear()
        # then
        assert len(sfs._ids) == 0

    def test_clear_empties_empty_buffer(self):
        # given
        sfs = collections.SnowflakeSet()
        assert len(sfs._ids) == 0
        # when
        sfs.clear()
        # then
        assert len(sfs._ids) == 0

    @pytest.mark.parametrize(
        ("start_with", "discard", "expect"),
        [
            ([], [1], []),
            ([1], [1], []),
            ([0], [1], [0]),
            ([0, 1], [1], [0]),
            ([0, 1], [0], [1]),
            ([9, 18, 27, 36, 45, 54, 63], [18, 27, 18, 18, 36, 64, 63], [9, 45, 54]),
        ],
    )
    def test_discard(self, start_with, discard, expect):
        # given
        sfs = collections.SnowflakeSet()
        sfs._ids.extend(start_with)
        # when
        for item in discard:
            sfs.discard(item)
        # then
        assert sfs._ids.tolist() == expect

    @pytest.mark.parametrize(
        ("start_with", "look_for", "expect"),
        [
            ([], 12, False),
            ([12], 12, True),
            ([9, 18, 27, 36, 45, 54, 63], 9, True),
            ([9, 18, 27, 36, 45, 54, 63], 18, True),
            ([9, 18, 27, 36, 45, 54, 63], 27, True),
            ([9, 18, 27, 36, 45, 54, 63], 36, True),
            ([9, 18, 27, 36, 45, 54, 63], 45, True),
            ([9, 18, 27, 36, 45, 54, 63], 54, True),
            ([9, 18, 27, 36, 45, 54, 63], 62, False),
            ([9, 18, 27, 36, 45, 54, 63], 63, True),
            ([9, 18, 27, 36, 45, 54, 63], 64, False),
            ([9, 18, 27, 36, 45, 54, 63], 72, False),
            ([12], "12", False),
        ],
    )
    def test_contains(self, start_with, look_for, expect):
        # given
        sfs = collections.SnowflakeSet()
        sfs._ids.extend(start_with)
        # then
        assert (look_for in sfs) is expect

    def test_iter(self):
        # given
        sfs = collections.SnowflakeSet()
        sfs._ids.extend([9, 18, 27, 36, 45, 54, 63])
        # then
        assert list(iter(sfs)) == [9, 18, 27, 36, 45, 54, 63]
        assert list(sfs) == [9, 18, 27, 36, 45, 54, 63]

    @pytest.mark.parametrize("items", [*range(0, 10)])
    def test_len(self, items):
        # given
        sfs = collections.SnowflakeSet()
        sfs._ids.extend(i for i in range(items))
        # then
        assert len(sfs) == items

    @pytest.mark.parametrize("items", [*range(0, 10)])
    def test_length_hint(self, items):
        # given
        sfs = collections.SnowflakeSet()
        sfs._ids.extend(i for i in range(items))
        # then
        assert operator.length_hint(sfs) == items

    def test_repr(self):
        sfs = collections.SnowflakeSet()
        sfs._ids.extend([1, 3, 5, 7, 8])
        assert repr(sfs) == "SnowflakeSet(1, 3, 5, 7, 8)"

    def test_str(self):
        sfs = collections.SnowflakeSet()
        sfs._ids.extend([1, 3, 5, 7, 8])
        assert str(sfs) == "{1, 3, 5, 7, 8}"

    def test_sizeof(self):
        sfs = collections.SnowflakeSet()
        sfs._ids.extend([1, 3, 5, 7, 8])
        assert sfs.__sizeof__() == object.__sizeof__(sfs) + sys.getsizeof(sfs._ids)


def test_get_index_or_slice_with_index_within_range():
    result = collections.get_index_or_slice({"i": "e", "n": "o", "b": "a", "hikari": "Rei", "p": "a", "o": "o"}, 3)
    assert result == "Rei"


def test_get_index_or_slice_with_index_outside_range():
    with pytest.raises(IndexError):
        collections.get_index_or_slice({"i": "e", "n": "o", "b": "a", "hikari": "noa"}, 77)


def test_get_index_or_slice_with_slice():
    test_map = {"o": "b", "b": "o", "a": "m", "arara": "blam", "oof": "no", "rika": "may"}
    assert collections.get_index_or_slice(test_map, slice(1, 5, 2)) == ("o", "blam")



# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import warnings

import mock
import pytest

from hikari import _about as hikari_about
from hikari.internal import deprecation


class TestWarnDeprecated:
    def test_when_not_past_removal(self):
        with mock.patch.object(hikari_about, "__version__", "2.0.1"):
            with mock.patch.object(warnings, "warn") as warn:
                deprecation.warn_deprecated(
                    "testing", removal_version="2.0.2", additional_info="Some info!", stack_level=100
                )

        warn.assert_called_once_with(
            "'testing' is deprecated and will be removed in `2.0.2`. Some info!",
            category=DeprecationWarning,
            stacklevel=100,
        )

    def test_when_past_removal(self):
        with mock.patch.object(hikari_about, "__version__", "2.0.2"):
            with pytest.raises(DeprecationWarning):
                deprecation.warn_deprecated("testing", removal_version="2.0.2", additional_info="Some info!")



# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import mock
import pytest

from hikari import snowflakes
from hikari.events import shard_events


class TestShardReadyEvent:
    @pytest.fixture
    def event(self):
        return shard_events.ShardReadyEvent(
            my_user=mock.Mock(),
            resume_gateway_url="testing",
            shard=None,
            actual_gateway_version=1,
            session_id="ok",
            application_id=1,
            application_flags=1,
            unavailable_guilds=[],
        )

    def test_app_property(self, event):
        assert event.app is event.my_user.app


class TestMemberChunkEvent:
    @pytest.fixture
    def event(self):
        return shard_events.MemberChunkEvent(
            app=mock.Mock(),
            shard=mock.Mock(),
            guild_id=snowflakes.Snowflake(69420),
            members={
                snowflakes.Snowflake(1): mock.Mock(),
                snowflakes.Snowflake(55): mock.Mock(),
                snowflakes.Snowflake(99): mock.Mock(),
                snowflakes.Snowflake(455): mock.Mock(),
            },
            chunk_count=1,
            chunk_index=1,
            not_found=(),
            presences={},
            nonce="blah",
        )

    def test___getitem___with_slice(self, event):
        mock_member_0 = object()
        mock_member_1 = object()
        event.members = {1: object(), 55: object(), 99: mock_member_0, 455: object(), 5444: mock_member_1}

        assert event[2:5:2] == (mock_member_0, mock_member_1)

    def test___getitem___with_valid_index(self, event):
        mock_member = object()
        event.members[snowflakes.Snowflake(99)] = mock_member
        assert event[2] is mock_member

        with pytest.raises(IndexError):
            assert event[55]

    def test___getitem___with_invalid_index(self, event):
        with pytest.raises(IndexError):
            assert event[123]

    def test___iter___(self, event):
        member_0 = mock.Mock()
        member_1 = mock.Mock()
        member_2 = mock.Mock()

        event.members = {
            snowflakes.Snowflake(1): member_0,
            snowflakes.Snowflake(2): member_1,
            snowflakes.Snowflake(3): member_2,
        }

        assert list(event) == [member_0, member_1, member_2]

    def test___len___(self, event):
        assert len(event) == 4
